unit UEngine;

{$IFDEF DARWIN}{$modeswitch Objectivec1}{$ENDIF}

interface

uses
  {$IFDEF DARWIN}MacOSAll, CocoaAll, CocoaUtils,{$ENDIF}
  Classes, SysUtils, Graphics, Controls, ExtCtrls, Math, DateUtils, Types,
  URegistry, UMetaCanvas, Generics.Collections, Generics.Defaults, UMenu,
  UControls, UMetroPassenger, UColors, UView, URiver, UTrack, UCity, UGeometric,
  UTranslator, DOM, XMLRead, XMLWrite, UXMLUtils;

type
  TEngine = class;
  TMetroLines = class;
  TMetroLine = class;
  TMetroTrains = class;
  TLineStation = class;
  TMetroTrain = class;

  { TMapStation }

  TMapStation = class
  private
    procedure ShiftTrackPoints;
    procedure SortLines;
  public
    Engine: TEngine;
    Shape: TStationShape;
    Position: TPoint;
    Passengers: TMetroPassengers;
    Lines: TMetroLines;
    ShapeDistance: array[TStationShape] of Integer;
    OverloadDuration: TDateTime;
    IsTerminal: Boolean;
    function GetMaxPassengers: Integer;
    function IsBestStationForShape(Shape: TStationShape; Check, Current: TLineStation): Boolean;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    constructor Create;
    destructor Destroy; override;
  end;

  { TMapStations }

  TMapStations = class(TObjectList<TMapStation>)
    Engine: TEngine;
    function GetRect: TRect;
    function AddNew: TMapStation;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
  end;

  { TLineStation }

  TLineStation = class
    Line: TMetroLine;
    MapStation: TMapStation;
    TrackPoint: TTrackPoint;
  end;

  { TLineStations }

  TLineStations = class(TObjectList<TLineStation>)
    Line: TMetroLine;
    function SearchMapStation(Station: TMapStation): TLineStation;
  end;

  { TMetroLine }

  TMetroLine = class
  private
    procedure UpdateEndingLine(EndIndex, Direction: Integer);
    procedure UpdateEndingLines;
  public
    Id: Integer;
    Index: Integer;
    Engine: TEngine;
    Color: TColor;
    LineStations: TLineStations;
    Trains: TMetroTrains;
    Track: TTrack;
    procedure ConnectStation(Station: TMapStation; LineStationDown, LineStationUp: TLineStation);
    procedure DisconnectStation(ALineStation: TLineStation);
    constructor Create;
    destructor Destroy; override;
    function IsCircular: Boolean;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
  end;

  { TMetroLines }

  TMetroLines = class(TObjectList<TMetroLine>)
    Engine: TEngine;
    function AddNew(Color: TColor): TMetroLine;
    function FindById(Id: Integer): TMetroLine;
    function SearchByColor(Color: TColor): TMetroLine;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
  end;

  { TMetroCarriage }

  TMetroCarriage = class
    Train: TMetroTrain;
    Passengers: TMetroPassengers;
    function GetTrackPosition: TTrackPosition;
    function GetVector: TVector;
    constructor Create;
    destructor Destroy; override;
  end;

  { TMetroCarriages }

  TMetroCarriages = class(TObjectList<TMetroCarriage>)
    function GetUnused: TMetroCarriage;
    function GetUnusedCount: Integer;
    function AddNew: TMetroCarriage;
  end;

  { TMetroTrain }

  TMetroTrain = class
  private
    FLine: TMetroLine;
    LastPosDelta: Integer;
    LastTrainMoveTime: TDateTime;
    StationStopTime: TDateTime;
    procedure SetLine(AValue: TMetroLine);
  public
    Engine: TEngine;
    Passengers: TMetroPassengers;
    TrackPosition: TTrackPosition;
    Direction: Integer;
    InStation: Boolean;
    TargetStation: TLineStation;
    Carriages: TMetroCarriages;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
    procedure FindTargetStation;
    function GetTargetStationDistance: Integer;
    constructor Create;
    destructor Destroy; override;
    property Line: TMetroLine read FLine write SetLine;
  end;

  { TMetroTrains }

  TMetroTrains = class(TObjectList<TMetroTrain>)
    function GetUnused: TMetroTrain;
    function GetUnusedCount: Integer;
    function AddNew: TMetroTrain;
    procedure LoadFromXmlNode(Node: TDOMNode);
    procedure SaveToXmlNode(Node: TDOMNode);
  end;

  { TMap }

  TMap = class
    Size: TPoint;
    Rivers: TRivers;
    constructor Create;
    destructor Destroy; override;
  end;

  TGameState = (gsNotStarted, gsRunning, gsPaused, gsGameOver, gsMenu, gsNewWeek,
    gsNewImprovement, gsSuccess);

  TMetroImprovement = (miNone, miTunnel, miTerminal, miLine, miCarriage,
    miFastTrain);
  TMetroImprovementSet = set of TMetroImprovement;
  TFullScrenChangeEvent = procedure (Sender: TObject; Active: Boolean) of object;

  { TEngine }

  TEngine = class(TComponent)
  private
    FDarkMode: Boolean;
    FFullScreen: Boolean;
    FOnDarkModeChange: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnFullScreenChange: TFullScrenChangeEvent;
    FState: TGameState;
    FTranslator: TTranslator;
    LastMousePos: TPoint;
    LastFocusedStation: TMapStation;
    MouseHold: Boolean;
    LastNewStationTime: TDateTime;
    LastNewPassengerTime: TDateTime;
    LastNewWeekTime: TDateTime;
    LastNewShapeTime: TDateTime;
    LastTickTime: TDateTime;
    FTime: TDateTime;
    FLastTime: TDateTime;
    MetaCanvas: TMetaCanvas;
    Menu: TMenu;
    MenuMain: TMenu;
    MenuOptions: TMenu;
    MenuCareer: TMenu;
    MenuGame: TMenu;
    LastState: TGameState;
    TimePerSecond: TDateTime;
    ImprovementImageSize: Integer;
    IconSize: Integer;
    LineColors: array of TColor;
    CarriageCountVisible: Boolean;
    TerminalCountVisible: Boolean;
    LastGrabbedTrainTrackPosition: TTrackPosition;
    GrabbedTrainDirection: Integer;
    LastGrabbedTrain: TMetroTrain;
    function GetServedDaysCount: Integer;
    procedure ResizeView(Force: Boolean);
    function GetExistStationShapes: TStationShapeSet;
    function GetStationOnPos(Pos: TPoint): TMapStation;
    function GetTrackOnPos(Pos: TPoint; out Intersect: TPoint): TTrackLink;
    function GetTrainOnPos(Pos: TPoint): TMetroTrain;
    function GetCarriageOnPos(Pos: TPoint): TMetroCarriage;
    procedure DrawFrame(Canvas: TCanvas; Rect: TRect);
    procedure DrawLine(Canvas: TCanvas; Pos: TPoint);
    procedure DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
      Size: Integer; Angle: Double);
    procedure DrawClock(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawTrains(Canvas: TCanvas);
    procedure DrawGameOver(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawHighScore(Canvas: TCanvas; CanvasSize: TPoint; Y: Integer);
    procedure DrawSuccess(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawNewWeek(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawNewImprovement(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawStationPassengerOverload(Canvas: TCanvas);
    procedure DrawLines(Canvas: TCanvas);
    procedure DrawStations(Canvas: TCanvas);
    procedure DrawGameControls(Canvas: TCanvas; CanvasSize: TPoint);
    procedure DrawGrabbed(Canvas: TCanvas; CanvasSize: TPoint);
    procedure ComputeShapeDistance;
    procedure ComputeShapeDistanceStation(Station: TMapStation;
      UpdatedShape: TStationShape; Distance: Integer);
    procedure SetDarkMode(AValue: Boolean);
    procedure SetFullScreen(AValue: Boolean);
    procedure SetState(AValue: TGameState);
    procedure TrainsMovement;
    procedure TrainMovement(Train: TMetroTrain);
    function GetUnusedLine: TMetroLine;
    procedure ShiftTrackPoints;
    procedure MenuItemExit(Sender: TObject);
    procedure MenuItemCity(Sender: TObject);
    procedure MenuItemPlay(Sender: TObject);
    procedure MenuItemCareer(Sender: TObject);
    procedure MenuItemOptions(Sender: TObject);
    procedure MenuItemGameContinue(Sender: TObject);
    procedure MenuItemGameExit(Sender: TObject);
    procedure MenuItemGameRestart(Sender: TObject);
    procedure MenuItemBack(Sender: TObject);
    procedure ButtonPlay(Sender: TObject);
    procedure ButtonPause(Sender: TObject);
    procedure ButtonFastForward(Sender: TObject);
    procedure ButtonNewTrain(Sender: TObject);
    procedure ButtonNewImprovement1(Sender: TObject);
    procedure ButtonNewImprovement2(Sender: TObject);
    procedure ButtonBackClick(Sender: TObject);
    procedure DarkModeChanged(Sender: TObject);
    procedure LanguageChanged(Sender: TObject);
    procedure FullScreenChanged(Sender: TObject);
    procedure UpdateInterface;
    function GetImprovementText(Improvement: TMetroImprovement): string;
    function GetImprovementBitmap(Improvement: TMetroImprovement): TBitmap;
    procedure EvaluateImprovement(Improvement: TMetroImprovement);
  public
    // Test
    function GetSelectedOrUnusedMetroLine: TMetroLine;
  public
    AvailableTerminals: Integer;
    Week: Integer;
    Colors: TColors;
    Passengers: TMetroPassengers;
    Stations: TMapStations;
    Lines: TMetroLines;
    Trains: TMetroTrains;
    Carriages: TMetroCarriages;
    ShapeCount: Integer;
    Map: TMap;
    View: TView;
    Cities: TCities;
    City: TCity;
    SelectedLine: TMetroLine;
    SelectedTrain: TMetroTrain;
    SelectedCarriage: TMetroCarriage;
    SelectedTerminal: Boolean;
    TrackStationDown: TTrackPoint;
    TrackStationUp: TTrackPoint;
    ServedPassengerCount: Integer;
    RedrawPending: Boolean;
    ButtonBack: TImage;
    ImagePassenger: TImage;
    ImageLocomotive: TImage;
    ImagePlay: TImage;
    ImagePause: TImage;
    ImageFastForward: TImage;
    ImageAchievement: TImage;
    ImageCarriage: TImage;
    ImageTerminal: TImage;
    ImageTunnel: TImage;
    ImageLine: TImage;
    ImageNewTrain: TImage;
    ImageNewImprovement1: TImage;
    ImageNewImprovement2: TImage;
    AvailableImprovements: TMetroImprovementSet;
    Improvement1: TMetroImprovement;
    Improvement2: TMetroImprovement;
    HighestServedPassengerCount: Integer;
    HighestServedDaysCount: Integer;
    RegistryContext: TRegistryContext;
    procedure InitMenus;
    procedure InitCities;
    procedure MouseMove(Position: TPoint);
    procedure MouseUp(Button: TMouseButton; Position: TPoint);
    procedure MouseDown(Button: TMouseButton; Position: TPoint);
    procedure KeyUp(Key: Word);
    procedure MainMenu;
    procedure Clear;
    procedure NewGame;
    procedure Redraw;
    procedure LoadFromRegistry;
    procedure SaveToRegistry;
    procedure LoadFromFile(FileName: string);
    procedure SaveToFile(FileName: string);
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Tick;
    procedure Paint(Canvas: TCanvas; CanvasSize: TPoint);
    property Time: TDateTime read FTime;
    property State: TGameState read FState write SetState;
    property ServedDaysCount: Integer read GetServedDaysCount;
  published
    property DarkMode: Boolean read FDarkMode write SetDarkMode;
    property FullScreen: Boolean read FFullScreen write SetFullScreen;
    property Translator: TTranslator read FTranslator write FTranslator;
    property OnDarkModeChange: TNotifyEvent read FOnDarkModeChange
      write FOnDarkModeChange;
    property OnFullScreenChange: TFullScrenChangeEvent read FOnFullScreenChange
      write FOnFullScreenChange;
    property OnExit: TNotifyEvent read FOnExit write FOnExit;
  end;

const
  StationSize = 30;
  StationOverloadSize = 60;
  PassengerSize = 15;
  TrainSize = 40;
  TrainGap = 5;
  LineColorsDist = 50;
  TrainSpeed = 2000;
  TrainPassengerCount = 6;
  StationMinDistance = 100;
  StationMaxDistance = 300;
  MaxWaitingPassengers = 10;
  MaxWaitingPassengersTerminal = 16;
  MaxPassengersOveloadTime = 2;
  MetroLineThickness = 13;
  TrackClickDistance = 15;
  EndStationLength = 50;
  ShowDistances = False;
  TimePerSecondNormal = 60 * OneMinute;
  TimePerSecondFast = 2 * TimePerSecondNormal;
  NewStationPeriod = 1;
  NewShapePeriod = 10;
  NewTrainPeriod = 7; // Each week
  NewPassengerPeriod = 0.3 * OneSecond;
  NewPassengerProbability = 0.003;
  VisiblePassengersPerLine = 6;


implementation

uses
  ULanguages;

resourcestring
  SGameOver = 'Game Over';
  SGameOverReason = 'Overcrowding at one of your stations has forced you to resign as metro manager.';
  SGameOverStatistic = '%d passengers travelled on your metro over %d days.';
  SDay = 'Day';
  SNewHighScore = 'New high score!';
  SOldHighScore = 'Old high score was %d passengers in %d days.';
  SStationNotDefined = 'Station have to be defined';
  SNoOldStationToConnectNew = 'No old line station to connect new station';
  SStationWithoutMapStation = 'Station have to have MapStation';
  SNewTrain = 'You get a new train for your metro';
  SNewImprovement = 'Select a new improvement for your metro';
  SWeek = 'week';
  STrain = 'Train';
  SPlay = 'Play';
  SOptions = 'Options';
  SExit = 'Exit';
  SBigMetro = 'Big Metro';
  SDarkMode = 'Dark mode';
  SLanguage = 'Language';
  SCzech = 'Czech';
  SEnglish = 'English';
  SFrench = 'French';
  SGerman = 'German';
  SAutomatic = 'Automatic';
  SBack = 'Back';
  SFullScreen = 'Full screen';
  SContinue = 'Continue';
  SRestart = 'Try again';
  STerminal = 'Terminal';
  SLine = 'Line';
  SCarriage = 'Carriage';
  STunnel = 'Tunnel';
  SCareer = 'Career';
  SSuccess = 'Success';
  SSuccessReason = 'Your metro transported enough passengers and you have reached your goal for this city.';
  SUnlockedCity = 'City %s is now unlocked.';
  SWrongFileFormat = 'Wrong file format';

  // Cities
  SPrague = 'Prague';
  SLondon = 'London';
  SParis = 'Paris';
  SNewYork = 'New York';
  STokyo = 'Tokyo';
  SRome = 'Rome';
  SSeoul = 'Seoul';
  SBeijing = 'Beijing';

{ TMetroCarriage }

function TMetroCarriage.GetTrackPosition: TTrackPosition;
begin
  if Assigned(Train) then begin
    Result := Train.TrackPosition;
    Result.Move(-Train.Direction * (TrainSize + TrainGap) * (Train.Carriages.IndexOf(Self) + 1));
  end;
end;

function TMetroCarriage.GetVector: TVector;
begin
  Result := Train.TrackPosition.GetVector;
  Result.Position := AddPoint(Result.Position, Point(TrainSize, TrainSize));
end;

constructor TMetroCarriage.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
end;

destructor TMetroCarriage.Destroy;
begin
  FreeAndNil(Passengers);
  inherited;
end;

{ TMetroCarriages }

function TMetroCarriages.GetUnused: TMetroCarriage;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Assigned(Items[I].Train)) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

function TMetroCarriages.GetUnusedCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Count - 1 do
    if not Assigned(Items[I].Train) then Inc(Result);
end;

function TMetroCarriages.AddNew: TMetroCarriage;
begin
  Result := TMetroCarriage.Create;
  Add(Result);
end;

{ TMap }

constructor TMap.Create;
begin
  Rivers := TRivers.Create;
end;

destructor TMap.Destroy;
begin
  FreeAndNil(Rivers);
  inherited;
end;

{ TLineStations }

function TLineStations.SearchMapStation(Station: TMapStation): TLineStation;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Items[I].MapStation <> Station) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

{ TMetroTrains }

function TMetroTrains.GetUnused: TMetroTrain;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Assigned(Items[I].Line)) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

function TMetroTrains.GetUnusedCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Count - 1 do
    if not Assigned(Items[I].Line) then Inc(Result);
end;

function TMetroTrains.AddNew: TMetroTrain;
begin
  Result := TMetroTrain.Create;
  Add(Result);
end;

procedure TMetroTrains.LoadFromXmlNode(Node: TDOMNode);
var
  Node2: TDOMNode;
  NewItem: TMetroTrain;
begin
  Node2 := Node.FirstChild;
  while Assigned(Node2) and (Node2.NodeName = 'Train') do begin
    NewItem := TMetroTrain.Create;
    NewItem.LoadFromXmlNode(Node2);
    Add(NewItem);
    Node2 := Node2.NextSibling;
  end;
end;

procedure TMetroTrains.SaveToXmlNode(Node: TDOMNode);
var
  I: Integer;
  NewNode: TDOMNode;
begin
  for I := 0 to Count - 1 do begin;
    NewNode := Node.OwnerDocument.CreateElement('Train');
    Node.AppendChild(NewNode);
    Items[I].SaveToXmlNode(NewNode);
  end;
end;

{ TMapStations }

function TMapStations.GetRect: TRect;
var
  I: Integer;
begin
  if Count > 0 then begin
    with Items[0] do
      Result := Rect(Position.X, Position.Y, Position.X, Position.Y);
    for I := 1 to Count - 1 do
    with Items[I] do begin
      if Position.X < Result.Left then Result.Left := Position.X;
      if Position.X > Result.Right then Result.Right := Position.X;
      if Position.Y < Result.Top then Result.Top := Position.Y;
      if Position.Y > Result.Bottom then Result.Bottom := Position.Y;
    end;
  end else Result := Rect(0, 0, 0, 0);
end;

function TMapStations.AddNew: TMapStation;
var
  D: Integer;
  MinD: Integer;
  I: Integer;
  Pass: Integer;
  Angle: Double;
  L: Integer;
const
  Step = 20;
begin
  Result := TMapStation.Create;
  Result.Engine := Engine;
  Angle := Random * 2 * Pi;
  // Ensure minimum distance between stations
  Pass := 0;
  L := Step;
  repeat
    Result.Position := Point(Trunc(Engine.Map.Size.X / 2 + Cos(Angle) * L * 1.5),
      Trunc(Engine.Map.Size.Y / 2 + Sin(Angle) * L));
    MinD := High(Integer);
    for I := 0 to Engine.Stations.Count - 1 do begin
      D := Distance(Engine.Stations[I].Position, Result.Position);
      if D < MinD then MinD := D;
    end;
    Inc(Pass);
    L := L + StationMinDistance div 2;
  until (MinD > StationMinDistance) or
    (Pass > 1000) or (Engine.Stations.Count = 0);
  Result.Shape := TStationShape(Random(Integer(Engine.ShapeCount)));
  Add(Result);
  Engine.ComputeShapeDistance;
end;

procedure TMapStations.LoadFromXmlNode(Node: TDOMNode);
var
  Node2: TDOMNode;
  NewItem: TMapStation;
begin
  Node2 := Node.FirstChild;
  while Assigned(Node2) and (Node2.NodeName = 'Line') do begin
    NewItem := TMapStation.Create;
    NewItem.LoadFromXmlNode(Node2);
    Add(NewItem);
    Node2 := Node2.NextSibling;
  end;
end;

procedure TMapStations.SaveToXmlNode(Node: TDOMNode);
var
  I: Integer;
  NewNode: TDOMNode;
begin
  for I := 0 to Count - 1 do begin;
    NewNode := Node.OwnerDocument.CreateElement('Station');
    Node.AppendChild(NewNode);
    Items[I].SaveToXmlNode(NewNode);
  end;
end;

{ TMetroLines }

function TMetroLines.AddNew(Color: TColor): TMetroLine;
begin
  Result := TMetroLine.Create;
  Result.Engine := Engine;
  Result.Index := Count;
  Result.Color := Color;
  Add(Result);
end;

function TMetroLines.FindById(Id: Integer): TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Items[I].Id <> Id) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

function TMetroLines.SearchByColor(Color: TColor): TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Items[I].Color <> Color) do Inc(I);
  if I < Count then Result := Items[I]
    else Result := nil;
end;

procedure TMetroLines.LoadFromXmlNode(Node: TDOMNode);
var
  Node2: TDOMNode;
  NewItem: TMetroLine;
begin
  Node2 := Node.FirstChild;
  while Assigned(Node2) and (Node2.NodeName = 'Line') do begin
    NewItem := TMetroLine.Create;
    NewItem.LoadFromXmlNode(Node2);
    Add(NewItem);
    Node2 := Node2.NextSibling;
  end;
end;

procedure TMetroLines.SaveToXmlNode(Node: TDOMNode);
var
  I: Integer;
  NewNode: TDOMNode;
begin
  for I := 0 to Count - 1 do begin;
    NewNode := Node.OwnerDocument.CreateElement('Line');
    Node.AppendChild(NewNode);
    Items[I].SaveToXmlNode(NewNode);
  end;
end;

{ TMetroLine }

procedure TMetroLine.UpdateEndingLine(EndIndex, Direction: Integer);
var
  //Index: Integer;
  //NewTrackPoint: TTrackPoint;
  Angle: Double;
  EndPoint: TPoint;
begin
{  if Direction = 1 then Index := Track.Points.IndexOf(LineStations.Last.TrackPoint)
    else if Direction = -1 then Index := Track.Points.IndexOf(LineStations.Last.TrackPoint);
  if Index = EndIndex then begin
    NewTrackPoint := Track.Points.AddNew;
    if Direction = 1 then Track.Points.Insert(EndIndex, NewTrackPoint)
    else if Direction = -1 then begin
      Inc(EndIndex);
      Track.Points.Insert(EndIndex, NewTrackPoint);
    end;
  end;
 }
  Angle := ArcTan2((Track.Points[EndIndex + 2 * Direction].PositionDesigned.Y -
    Track.Points[EndIndex + Direction].PositionDesigned.Y),
    (Track.Points[EndIndex + 2 * Direction].PositionDesigned.X -
    Track.Points[EndIndex + Direction].PositionDesigned.X));
  EndPoint := Point(Round(Track.Points[EndIndex + Direction].PositionDesigned.X - EndStationLength * Cos(Angle)),
    Round(Track.Points[EndIndex + Direction].PositionDesigned.Y - EndStationLength * Sin(Angle)));
  Track.Points[EndIndex].PositionDesigned := EndPoint;
  Track.Points[EndIndex].Position := EndPoint;
end;

procedure TMetroLine.UpdateEndingLines;
var
  Index: Integer;
  NewTrackPoint: TTrackPoint;
begin
  if LineStations.Count >= 2 then begin
    Index := Track.Points.IndexOf(LineStations.First.TrackPoint);
    if Index = 0 then begin
      NewTrackPoint := Track.Points.AddNew;
      Track.Points.Insert(0, NewTrackPoint);
    end;
    Index := Track.Points.IndexOf(LineStations.Last.TrackPoint);
    if Index = Track.Points.Count - 1 then begin
      NewTrackPoint := Track.Points.AddNew;
      Track.Points.Insert(Track.Points.Count, NewTrackPoint);
    end;
    UpdateEndingLine(0, 1);
    UpdateEndingLine(Track.Points.Count - 1, -1);
  end;
end;

procedure TMetroLine.ConnectStation(Station: TMapStation; LineStationDown, LineStationUp: TLineStation);
var
  Train: TMetroTrain;
  NewTrackPoint: TTrackPoint;
  NewLineStation: TLineStation;
  Index: Integer;
begin
  if not Assigned(Station) then
    raise Exception.Create(SStationNotDefined);
  if not Assigned(LineStationDown) and not Assigned(LineStationUp) and (LineStations.Count > 0) then
    raise Exception.Create(SNoOldStationToConnectNew);
  NewLineStation := TLineStation.Create;
  NewLineStation.Line := Self;
  NewLineStation.MapStation := Station;
  Index := 0;
  if Assigned(LineStationDown) then Index := LineStations.IndexOf(LineStationDown) + 1
    else if Assigned(LineStationDown) then Index := LineStations.IndexOf(LineStationUp);
  LineStations.Insert(Index, NewLineStation);
  Station.Lines.Add(Self);

  NewTrackPoint := Track.Points.AddNew;
  NewTrackPoint.OwnerPoint := NewLineStation;
  NewTrackPoint.Position := Station.Position;
  NewTrackPoint.PositionDesigned := NewTrackPoint.Position;
  Index := 0;
  if Assigned(LineStationDown) then Index := Track.Points.IndexOf(LineStationDown.TrackPoint) + 1
    else if Assigned(LineStationUp) then Index := Track.Points.IndexOf(LineStationUp.TrackPoint);
  Track.Points.Insert(Index, NewTrackPoint);
  NewLineStation.TrackPoint := NewTrackPoint;

  if Assigned(LineStationDown) then
    Track.RouteTrack(NewLineStation.TrackPoint.GetDown, NewLineStation.TrackPoint);
  if Assigned(LineStationUp) then
    Track.RouteTrack(NewLineStation.TrackPoint, NewLineStation.TrackPoint.GetUp);

  // Place one train if at least two stations present
  if (LineStations.Count = 2) then begin
    Train := Engine.Trains.GetUnused;
    if Assigned(Train) then begin
      Train.Line := Self;
      Train.TargetStation := LineStations[0];
      Train.TrackPosition.BaseTrackPoint := Track.Points.First;
      Trains.Add(Train);
    end;
  end;
  UpdateEndingLines;
  Engine.ComputeShapeDistance;
  Engine.ShiftTrackPoints;
end;

procedure TMetroLine.DisconnectStation(ALineStation: TLineStation);
var
  I: Integer;
  J: Integer;
  Index: Integer;
  TP1, TP2: TTrackPoint;
  IsOnTrack: Boolean;
begin
  // Determine track point range to be removed
  TP1 := ALineStation.TrackPoint.GetDown;
  if not Assigned(TP1) then TP1 := Track.Points.First;
  TP2 := ALineStation.TrackPoint.GetUp;
  if not Assigned(TP2) then TP2 := Track.Points.Last;

  // Remove track points from trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    IsOnTrack := False;
    for J := Track.Points.IndexOf(TP1) to Track.Points.IndexOf(TP2) do
    if Track.Points[J] = TrackPosition.BaseTrackPoint then begin
      IsOnTrack := True;
      Break;
    end;
    if IsOnTrack then begin
      if Assigned(TrackPosition.BaseTrackPoint) and Assigned(TrackPosition.BaseTrackPoint.GetUp) and
        (TrackPosition.BaseTrackPoint.GetUp <> ALineStation.TrackPoint) then
        TrackPosition.BaseTrackPoint := TrackPosition.BaseTrackPoint.GetUp
      else
      if Assigned(TrackPosition.BaseTrackPoint) and Assigned(TrackPosition.BaseTrackPoint.GetDown) and
        (TrackPosition.BaseTrackPoint.GetDown <> ALineStation.TrackPoint) then
        TrackPosition.BaseTrackPoint := TrackPosition.BaseTrackPoint.GetDown
      else TrackPosition.BaseTrackPoint := nil;
    end;
  end;

  // Delete old trackpoints
  Index := Track.Points.IndexOf(ALineStation.TrackPoint) - 1;
  while (Index >= 0) and (not Assigned(Track.Points[Index].OwnerPoint)) do begin
    Track.Points.Delete(Index);
    Dec(Index);
  end;
  Index := Index + 1;
  Track.Points.Delete(Index);
  while (Index < Track.Points.Count) and (not Assigned(Track.Points[Index].OwnerPoint)) do
    Track.Points.Delete(Index);

  if ((Index - 1) >= 0) and (Index < Track.Points.Count) then
    Track.RouteTrack(Track.Points[Index - 1], Track.Points[Index]);

  ALineStation.MapStation.Lines.Remove(Self);
  Index := LineStations.IndexOf(ALineStation);

  for I := 0 to Trains.Count - 1 do
  with Trains[I] do begin
    if TargetStation = ALineStation then
      TargetStation := LineStations[(Index + 1) mod LineStations.Count];
  end;

  LineStations.Delete(Index);

  // Remove all trains if less then two stations
  if LineStations.Count < 2 then
  for I := Trains.Count - 1 downto 0 do begin
    Trains[I].Line := nil;
    Trains.Delete(I);
  end;
  UpdateEndingLines;
  Engine.ComputeShapeDistance;
  Engine.ShiftTrackPoints;
end;

constructor TMetroLine.Create;
begin
  LineStations := TLineStations.Create;
  LineStations.OwnsObjects := True;
  Trains := TMetroTrains.Create;
  Trains.OwnsObjects := False;
  Track := TTrack.Create;
  Track.Owner := Self;
end;

destructor TMetroLine.Destroy;
begin
  FreeAndNil(Trains);
  FreeAndNil(LineStations);
  FreeAndNil(Track);
  inherited;
end;

function TMetroLine.IsCircular: Boolean;
begin
  Result := False;
  if LineStations.Count >= 2 then
    Result := (LineStations.Last.MapStation = LineStations.First.MapStation);
end;

procedure TMetroLine.LoadFromXmlNode(Node: TDOMNode);
var
  Node2: TDOMNode;
begin
  Id := ReadInteger(Node, 'Id', Id);
  Color := TColor(ReadInteger(Node, 'Color', Color));

  Node2 := Node.FindNode('Track');
  if Assigned(Node2) then
    Track.LoadFromXmlNode(Node2);
end;

procedure TMetroLine.SaveToXmlNode(Node: TDOMNode);
var
  Node2: TDOMNode;
begin
  WriteInteger(Node, 'Id', Id);
  WriteInteger(Node, 'Color', Color);

  Node2 := Node.OwnerDocument.CreateElement('Track');
  Node.AppendChild(Node2);
  Track.SaveToXmlNode(Node2);
end;

{ TMetroTrain }

procedure TMetroTrain.SetLine(AValue: TMetroLine);
begin
  if FLine = AValue then Exit;
  FLine := AValue;
  if AValue = nil then begin
    TrackPosition.RelPos := 0;
    TrackPosition.BaseTrackPoint := nil;
    TargetStation := nil;
  end;
end;

procedure TMetroTrain.LoadFromXmlNode(Node: TDOMNode);
begin
  Line := Engine.Lines.FindById(ReadInteger(Node, 'Line', 0));
end;

procedure TMetroTrain.SaveToXmlNode(Node: TDOMNode);
begin
  if Assigned(Line) then WriteInteger(Node, 'Line', Line.Id)
    else WriteInteger(Node, 'Line', 0);
end;

procedure TMetroTrain.FindTargetStation;
var
  TP: TTrackPoint;
begin
  if Direction > 0 then begin
    TP := TrackPosition.BaseTrackPoint.GetUp;
    if Assigned(TP) then begin
      TargetStation := TLineStation(TP.OwnerPoint);
    end else begin
      TP := TrackPosition.BaseTrackPoint.GetDown;
      if Assigned(TP) then
        TargetStation := TLineStation(TP.OwnerPoint);
    end;
  end else
  if Direction < 0 then begin
    if Assigned(TrackPosition.BaseTrackPoint.OwnerPoint) then
      TargetStation := TLineStation(TrackPosition.BaseTrackPoint.OwnerPoint)
    else begin
      TP := TrackPosition.BaseTrackPoint.GetUp;
      if Assigned(TP) then
        TargetStation := TLineStation(TP.OwnerPoint);
    end;
  end;
  LastPosDelta := Abs(GetTargetStationDistance);
end;

function TMetroTrain.GetTargetStationDistance: Integer;
var
  Current: Integer;
  Target: Integer;
  I: Integer;
begin
  Result := 0;
  if Assigned(TrackPosition.BaseTrackPoint) and Assigned(TargetStation) then begin
    Current := Line.Track.Points.IndexOf(TrackPosition.BaseTrackPoint);
    Target := Line.Track.Points.IndexOf(TargetStation.TrackPoint);
    if Current < Target then begin
      for I := Current to Target - 1 do
        Result := Result + Line.Track.Points[I].GetDistance;
      Result := Result - Trunc(TrackPosition.RelPos);
    end else
    if Current > Target then begin
      for I := Current - 1 downto Target do
        Result := Result + Line.Track.Points[I].GetDistance;
      Result := Result + Trunc(TrackPosition.RelPos);
    end else Result := Trunc(TrackPosition.RelPos);
  end;
end;

constructor TMetroTrain.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Carriages := TMetroCarriages.Create;
  Carriages.OwnsObjects := False;
  Direction := 1;
  Line := nil;
end;

destructor TMetroTrain.Destroy;
begin
  FreeAndNil(Passengers);
  FreeAndNil(Carriages);
  inherited;
end;

{ TMapStation }

procedure TMapStation.ShiftTrackPoints;
var
  TrackLinks: TTrackLinks;
  I: Integer;
  J: Integer;
  Index: Integer;
  TP: TTrackPoint;
  LS: TLineStation;
  Line: TMetroLine;
  Angle: Float;
  TPAngleGroup: TTrackPointsAngleGroup;
  GroupItem: TTrackPointsAngle;
  NewTrackLink: TTrackLink;
  HAngle: Double;
  P1, P2: TPoint;
  NewShift: TPoint;
begin
  TrackLinks := TTrackLinks.Create;
  TrackLinks.OwnsObjects := False;

  // Collect all near track points as track links
  SortLines;
  for I := 0 to Lines.Count - 1 do begin
    Line := Lines[I];
    LS := Line.LineStations.SearchMapStation(Self);
    TP := LS.TrackPoint;
    Index := Line.Track.Points.IndexOf(TP);
    if Index > 0 then begin
      NewTrackLink := Line.Track.Points[Index].GetLinkDown;
      TrackLinks.Add(NewTrackLink);
    end;
    if Index < (Line.Track.Points.Count - 1) then begin
      NewTrackLink := Line.Track.Points[Index].GetLinkUp;
      TrackLinks.Add(NewTrackLink);
    end;
    if Line.IsCircular and (Self = Line.LineStations.First.MapStation) and
    (Self = Line.LineStations.Last.MapStation) then begin
      LS := Line.LineStations.Last;
      TP := LS.TrackPoint;
      Index := Line.Track.Points.IndexOf(TP);
      if Index > 0 then begin
        NewTrackLink := Line.Track.Points[Index].GetLinkDown;
        TrackLinks.Add(NewTrackLink);
      end;
      if Index < (Line.Track.Points.Count - 1) then begin
        NewTrackLink := Line.Track.Points[Index].GetLinkUp;
        TrackLinks.Add(NewTrackLink);
      end;
    end;
  end;

  // Make groups of TrackLinks with same angle
  TPAngleGroup := TTrackPointsAngleGroup.Create;
  for I := 0 to TrackLinks.Count - 1 do begin
    P1 := TrackLinks[I].Points[0].PositionDesigned;
    P2 := TrackLinks[I].Points[1].PositionDesigned;
    if ComparePoint(P1, Position) and not ComparePoint(P2, Position) then begin
      Angle := ArcTan2(P2.Y - Position.Y, P2.X - Position.X);
    end else
    if ComparePoint(P2, Position) and not ComparePoint(P1, Position) then begin
      Angle := ArcTan2(P1.Y - Position.Y, P1.X - Position.X);
    end else Angle := 0;// else raise Exception.Create('TrackLink angle error');

    GroupItem := TPAngleGroup.SearchAngle(Angle);
    if not Assigned(GroupItem) then begin
      GroupItem := TTrackPointsAngle.Create;
      GroupItem.Angle := Angle;
      TPAngleGroup.Add(GroupItem);
    end;
    GroupItem.TrackLinks.Add(TrackLinks[I]);
  end;

  // Shift TrackLinks according number of lines in group
  for I := 0 to TPAngleGroup.Count - 1 do
  with TPAngleGroup[I] do begin
    for J := 0 to TrackLinks.Count - 1 do
      with TrackLinks[J] do begin
        // Get orthogonal angle
        HAngle := (Angle + Pi / 2) mod Pi;
        NewShift.X := Trunc(MetroLineThickness * Cos(HAngle) * (J - (TrackLinks.Count - 1) / 2));
        NewShift.Y := Trunc(MetroLineThickness * Sin(HAngle) * (J - (TrackLinks.Count - 1) / 2));
        Shift := NewShift;
      end;
  end;

  FreeAndNil(TPAngleGroup);
  FreeAndNil(TrackLinks);
end;

function MapStationCompareLine(constref Item1, Item2: TMetroLine): Integer;
begin
  if Item1.Index > Item2.Index then Result := 1
  else if Item1.Index < Item2.Index then Result := -1
  else Result := 0;
end;

procedure TMapStation.SortLines;
begin
  Lines.Sort(TComparer<TMetroLine>.Construct(MapStationCompareLine));
end;

function TMapStation.GetMaxPassengers: Integer;
begin
  if IsTerminal then Result := MaxWaitingPassengersTerminal
    else Result := MaxWaitingPassengers;
end;

function TMapStation.IsBestStationForShape(Shape: TStationShape;
  Check, Current: TLineStation): Boolean;
var
  I: Integer;
  T: Integer;
  Distance: Integer;
  StationIndex: Integer;
  DirectionUp: Boolean;
  DirectionDown: Boolean;
  NextStationUp: TLineStation;
  NextStationDown: TLineStation;
  CurrentLineStation: TLineStation;
begin
  Distance := High(Integer);
  for I := 0 to Lines.Count - 1 do
  with Lines[I] do begin
    CurrentLineStation := LineStations.SearchMapStation(Current.MapStation);
    StationIndex := LineStations.IndexOf(CurrentLineStation);
    if IsCircular then begin
      DirectionUp := False;
      DirectionDown := False;
      for T := 0 to Trains.Count - 1 do begin
        if Trains[T].Direction = 1 then DirectionUp := True;
        if Trains[T].Direction = -1 then DirectionDown := True;
      end;
      if StationIndex = 0 then
        NextStationDown := LineStations[LineStations.Count - 2]
      else
      if StationIndex > 0 then
        NextStationDown := LineStations[StationIndex - 1];

      if (StationIndex >= 0) and (StationIndex = LineStations.Count - 1) then
        NextStationUp := LineStations[1]
      else
      if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
        NextStationUp := LineStations[StationIndex + 1];
    end else begin
      if StationIndex > 0 then begin
        DirectionDown := True;
        NextStationDown := LineStations[StationIndex - 1]
      end else DirectionDown := False;
      if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then begin
        DirectionUp := True;
        NextStationUp := LineStations[StationIndex + 1];
      end else DirectionUp := False;
    end;
    if DirectionDown and (NextStationDown.MapStation.ShapeDistance[Shape] <> -1) and
      (NextStationDown.MapStation.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationDown.MapStation.ShapeDistance[Shape];
    end;
    if DirectionUp and (NextStationUp.MapStation.ShapeDistance[Shape] <> -1) and
      (NextStationUp.MapStation.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationUp.MapStation.ShapeDistance[Shape];
    end;
  end;
  Result := (Check.MapStation.ShapeDistance[Shape] <> -1) and
    (Check.MapStation.ShapeDistance[Shape] <= Distance);
end;

procedure TMapStation.LoadFromXmlNode(Node: TDOMNode);
begin
  Position.X := ReadInteger(Node, 'PositionX', Position.X);
  Position.Y := ReadInteger(Node, 'PositionY', Position.Y);
  Shape := TStationShape(ReadInteger(Node, 'Shape', Integer(Shape)));
  IsTerminal := ReadBoolean(Node, 'IsTerminal', IsTerminal);
end;

procedure TMapStation.SaveToXmlNode(Node: TDOMNode);
begin
  WriteInteger(Node, 'PositionX', Position.X);
  WriteInteger(Node, 'PositionY', Position.Y);
  WriteInteger(Node, 'Shape', Integer(Shape));
  WriteBoolean(Node, 'IsTerminal', IsTerminal);
end;

constructor TMapStation.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Lines := TMetroLines.Create;
  Lines.OwnsObjects := False;
end;

destructor TMapStation.Destroy;
begin
  FreeAndNil(Lines);
  FreeAndNil(Passengers);
  inherited;
end;

{ TEngine }

// Need to see all stations on screen
procedure TEngine.ResizeView(Force: Boolean);
var
  NewPoint: TPoint;
  Intersected: TRect;
  NewView: TView;
begin
  NewView := TView.Create;
  NewView.Assign(View);

  NewView.SourceRect := RectEnlarge(Stations.GetRect, 100);
  NewPoint := Point(
    Trunc((NewView.SourceRect.Left + (NewView.SourceRect.Right - NewView.SourceRect.Left) / 2) -
    (NewView.DestRect.Left + (NewView.DestRect.Right - NewView.DestRect.Left) / 2 / NewView.Zoom)),
    Trunc((NewView.SourceRect.Top + (NewView.SourceRect.Bottom - NewView.SourceRect.Top) / 2) -
    (NewView.DestRect.Top + (NewView.DestRect.Bottom - NewView.DestRect.Top) / 2 / NewView.Zoom)));
  NewView.SourceRect := Bounds(NewPoint.X, NewPoint.Y, Trunc((NewView.DestRect.Right - NewView.DestRect.Left) / NewView.Zoom),
    Trunc((NewView.DestRect.Bottom - NewView.DestRect.Top) / NewView.Zoom));

  Intersected := NewView.SourceRect;
  Intersected.Union(View.SourceRect);
  if not Force and (Intersected = View.SourceRect) then Exit;

  View.Assign(NewView);

  NewView.Free;
end;

function TEngine.GetServedDaysCount: Integer;
begin
  Result := Trunc(Time);
end;

function TEngine.GetExistStationShapes: TStationShapeSet;
var
  Station: TMapStation;
begin
  Result := [];
  for Station in Stations do
    Result := Result + [Station.Shape];
end;

function TEngine.GetStationOnPos(Pos: TPoint): TMapStation;
var
  I: Integer;
const
  ClickDistance = 30;
begin
  I := 0;
  while (I < Stations.Count) and (Distance(Stations[I].Position, Pos) > ClickDistance) do Inc(I);
  if I < Stations.Count then Result := Stations[I]
    else Result := nil;
end;

function TEngine.GetTrackOnPos(Pos: TPoint; out Intersect: TPoint): TTrackLink;
var
  I: Integer;
  T: Integer;
  D: Integer;
  MinD: Integer;
  TempIntersect: TPoint;
begin
  Result := TTrackLink.Create;
  Result.Points.Count := 2;
  Result.Points[0] := nil;
  Result.Points[1] := nil;
  I := 0;
  MinD := High(Integer);
  while (I < Lines.Count) do
  with TMetroLine(Lines[I]) do begin
    for T := 1 to Track.Points.Count - 1 do begin
      D := PointToLineDistance(Pos, Track.Points[T - 1].Position, Track.Points[T].Position,
        TempIntersect);
      if (D < MinD) and (D < TrackClickDistance) then begin
        MinD := D;
        Result.Points[0] := Track.Points[T - 1];
        Result.Points[1] := Track.Points[T];
        Intersect := TempIntersect;
      end;
    end;
    Inc(I);
  end;
end;

function TEngine.GetTrainOnPos(Pos: TPoint): TMetroTrain;
var
  I: Integer;
  MinDistance: Integer;
  D: Integer;
begin
  Result := nil;
  MinDistance := High(Integer);
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    D := Distance(TrackPosition.GetVector.Position, Pos);
    if (D < (TrainSize div 2)) and (D < MinDistance) then begin
      Result := Trains[I];
      MinDistance := D;
    end;
  end;
end;

function TEngine.GetCarriageOnPos(Pos: TPoint): TMetroCarriage;
var
  I: Integer;
  J: Integer;
  MinDistance: Integer;
  D: Integer;
begin
  Result := nil;
  MinDistance := High(Integer);
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    for J := 0 to Carriages.Count - 1 do
    with TMetroCarriage(Carriages[J]) do begin
      D := Distance(GetTrackPosition.GetVector.Position, Pos);
      if (D < (TrainSize div 2)) and (D < MinDistance) then begin
        Result := Carriages[J];
        MinDistance := D;
      end;
    end;
  end;
end;

procedure TEngine.DrawFrame(Canvas: TCanvas; Rect: TRect);
begin
  with Canvas do begin
    Pen.Color := Self.Colors.Text;
    Pen.Style := psSolid;
    Pen.Width := ScaleX(2, 96);
    Brush.Color := Self.Colors.Background2;
    Brush.Style := bsSolid;
    Rectangle(Rect);
    Pen.Style := psClear;
  end;
end;

procedure TEngine.DrawLine(Canvas: TCanvas; Pos: TPoint);
var
  Delta: TPoint;
begin
  Delta := Point(Pos.X - Canvas.PenPos.X, Pos.Y - Canvas.PenPos.Y);
  if Abs(Delta.X) > Abs(Delta.Y) then begin
    Canvas.LineTo(Pos.X - Sign(Delta.X) * Abs(Delta.Y), Canvas.PenPos.Y);
  end else begin
    Canvas.LineTo(Canvas.PenPos.X, Pos.Y - Sign(Delta.Y) * Abs(Delta.X));
  end;
  Canvas.LineTo(Pos.X, Pos.Y);
end;

procedure TEngine.DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
  Size: Integer; Angle: Double);
var
  Points: array of TPoint;
  I: Integer;
  Angle2: Double;
begin
  case Shape of
    ssSquare: begin
      SetLength(Points, 4);
      Points[0] := Point(Position.X - Size div 2, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 2);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 2);
      Points[3] := Point(Position.X - Size div 2, Position.Y + Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssCircle: Canvas.Ellipse(
      Position.X - Size div 2, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2);
    ssTriangle: begin
      SetLength(Points, 3);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y + Size div 2);
      Points[2] := Point(Position.X - Size div 2, Position.Y + Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssStar: begin
      SetLength(Points, 10);
      for I := 0 to 9 do begin
        Angle2 := I / 10 * 2 * Pi - Pi / 2;
        if (I mod 2) = 0 then
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2))
        else
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 5),
          Round(Position.Y + Sin(Angle2) * Size / 5));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssPlus: begin
      SetLength(Points, 12);
      Points[0] := Point(Position.X + Size div 6, Position.Y - Size div 6);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 6);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 6);
      Points[3] := Point(Position.X + Size div 6, Position.Y + Size div 6);
      Points[4] := Point(Position.X + Size div 6, Position.Y + Size div 2);
      Points[5] := Point(Position.X - Size div 6, Position.Y + Size div 2);
      Points[6] := Point(Position.X - Size div 6, Position.Y + Size div 6);
      Points[7] := Point(Position.X - Size div 2, Position.Y + Size div 6);
      Points[8] := Point(Position.X - Size div 2, Position.Y - Size div 6);
      Points[9] := Point(Position.X - Size div 6, Position.Y - Size div 6);
      Points[10] := Point(Position.X - Size div 6, Position.Y - Size div 2);
      Points[11] := Point(Position.X + Size div 6, Position.Y - Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssPentagon: begin
      SetLength(Points, 5);
      for I := 0 to 4 do begin
        Angle2 := I / 5 * 2 * Pi - Pi / 2;
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssHexagon: begin
      SetLength(Points, 6);
      for I := 0 to 5 do begin
        Angle2 := I / 6 * 2 * Pi - Pi / 2;
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssDiamond: begin
      SetLength(Points, 4);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y);
      Points[2] := Point(Position.X, Position.Y + Size div 2);
      Points[3] := Point(Position.X - Size div 2, Position.Y);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssCross: begin
      SetLength(Points, 12);
      Points[0] := Point(Position.X + Size div 6, Position.Y - Size div 6);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 6);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 6);
      Points[3] := Point(Position.X + Size div 6, Position.Y + Size div 6);
      Points[4] := Point(Position.X + Size div 6, Position.Y + Size div 2);
      Points[5] := Point(Position.X - Size div 6, Position.Y + Size div 2);
      Points[6] := Point(Position.X - Size div 6, Position.Y + Size div 6);
      Points[7] := Point(Position.X - Size div 2, Position.Y + Size div 6);
      Points[8] := Point(Position.X - Size div 2, Position.Y - Size div 6);
      Points[9] := Point(Position.X - Size div 6, Position.Y - Size div 6);
      Points[10] := Point(Position.X - Size div 6, Position.Y - Size div 2);
      Points[11] := Point(Position.X + Size div 6, Position.Y - Size div 2);
      Points := RotatePoints(Position, Points, Angle + Pi / 4);
      Canvas.Polygon(Points);
    end;
    ssHalfCircle: Canvas.Pie(
      Position.X - Size div 2, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2,
      Position.X - Size div 2, Position.Y,
      Position.X + Size div 2, Position.Y);
    ssQuarterCircle: Canvas.Pie(
      Position.X - Size div 2 - Size, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2 + Size,
      Position.X + Size div 2, Position.Y + Size div 2,
      Position.X - Size div 2, Position.Y - Size div 2);
    ssHeptagon: begin
      SetLength(Points, 8);
      for I := 0 to High(Points) do begin
        Angle2 := I / Length(Points) * 2 * Pi - Pi / 2;
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
  end;
end;

procedure TEngine.ComputeShapeDistance;
var
  S: TStationShape;
  Station: TMapStation;
begin
  // NewGame all distances
  for Station in Stations do
  with Station do begin
    for S := Low(ShapeDistance) to High(ShapeDistance) do
      ShapeDistance[S] := -1;
  end;

  // Propagate shape distance for all stations
  // Distace 0 means that station is final target
  for Station in Stations do
  with Station do begin
    ComputeShapeDistanceStation(Station, Shape, 0);
  end;
end;

procedure TEngine.ComputeShapeDistanceStation(Station: TMapStation;
  UpdatedShape: TStationShape; Distance: Integer);
var
  I: Integer;
  T: Integer;
  StationIndex: Integer;
  DirectionDown: Boolean;
  DirectionUp: Boolean;
begin
  with Station do begin
    if (Distance < ShapeDistance[UpdatedShape]) or (ShapeDistance[UpdatedShape] = -1) then begin
      ShapeDistance[UpdatedShape] := Distance;
      // Do for all lines connected to station
      for I := 0 to Lines.Count - 1 do
      with Lines[I] do
        for StationIndex := 0 to LineStations.Count - 1 do
        if LineStations[StationIndex].MapStation = Station then begin
        if not IsCircular then begin
          // Update for all adjecent stations
          if StationIndex > 0 then
            ComputeShapeDistanceStation(LineStations[StationIndex - 1].MapStation,
            UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
            ComputeShapeDistanceStation(LineStations[StationIndex + 1].MapStation,
            UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
        end else begin
          // If circular then trains might go in single direction so passengers
          // waiting for opposite directions are wrong
          DirectionUp := False;
          DirectionDown := False;
          for T := 0 to Trains.Count - 1 do begin
            if Trains[T].Direction = 1 then DirectionUp := True;
            if Trains[T].Direction = -1 then DirectionDown := True;
          end;
          // Update for all adjecent stations
          if DirectionUp then begin
            if StationIndex = 0 then
              ComputeShapeDistanceStation(LineStations[LineStations.Count - 2].MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if StationIndex > 0 then
              ComputeShapeDistanceStation(LineStations[StationIndex - 1].MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
          if DirectionDown then begin
            if (StationIndex >= 0) and (StationIndex = LineStations.Count - 1) then
              ComputeShapeDistanceStation(LineStations[1].MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
              ComputeShapeDistanceStation(LineStations[StationIndex + 1].MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
        end;
      end;
    end;
  end;
end;

procedure TEngine.SetDarkMode(AValue: Boolean);
begin
  if FDarkMode = AValue then Exit;
  FDarkMode := AValue;
  Colors.Init(FDarkMode);
  if Assigned(FOnDarkModeChange) then FOnDarkModeChange(Self);
end;

procedure TEngine.SetFullScreen(AValue: Boolean);
begin
  if FFullScreen = AValue then Exit;
  FFullScreen := AValue;
  if Assigned(FOnFullScreenChange) then
    FOnFullScreenChange(Self, FFullScreen);
end;

procedure TEngine.SetState(AValue: TGameState);
begin
  if FState = AValue then Exit;
  FState := AValue;
  UpdateInterface;
end;

procedure TEngine.TrainsMovement;
var
  I: Integer;
begin
  for I := 0 to Trains.Count - 1 do
    TrainMovement(Trains[I]);
end;

procedure TEngine.TrainMovement(Train: TMetroTrain);
var
  J: Integer;
  CurrentStation: TLineStation;
  P: Integer;
  Passenger: TMetroPassenger;
  PosDelta: Integer;
  TargetStationIndex: Integer;
  PosChange: Double;
  Done: Boolean;
begin
  with Train do begin
    if Assigned(Line) then begin
      if InStation then begin
        if (Time - StationStopTime) > OneHour then begin
          CurrentStation := TargetStation;

          // Choose next target station
          TargetStationIndex := Line.LineStations.IndexOf(TargetStation) + Direction;
          if TargetStationIndex < 0 then begin
            if Line.IsCircular then begin
              TargetStationIndex := Line.LineStations.Count - 2;
              TrackPosition.BaseTrackPoint := Line.LineStations.Last.TrackPoint;
              TrackPosition.RelPos := 0;
            end else begin
              TargetStationIndex := 1;
              Direction := -Direction;
            end;
          end else
          if TargetStationIndex >= Line.LineStations.Count then begin
            if Line.IsCircular then begin
              TargetStationIndex := 1;
              TrackPosition.BaseTrackPoint := Line.LineStations.First.TrackPoint;
              TrackPosition.RelPos := 0;
            end else begin
              TargetStationIndex := Line.LineStations.Count - 2;
              Direction := -Direction;
            end;
          end;
          TargetStation := Line.LineStations[TargetStationIndex];

          // Unload passengers in target station
          if Assigned(CurrentStation) then begin
            for P := Passengers.Count - 1 downto 0 do begin
              if Passengers[P].Shape = CurrentStation.MapStation.Shape then begin
                Passenger := Passengers[P];
                Passengers.Delete(P);
                Self.Passengers.Remove(Passenger);
                Inc(ServedPassengerCount);
              end;
            end;
            for J := 0 to Carriages.Count - 1 do
            with Carriages[J] do begin
              for P := Passengers.Count - 1 downto 0 do begin
                if Passengers[P].Shape = CurrentStation.MapStation.Shape then begin
                  Passenger := Passengers[P];
                  Passengers.Delete(P);
                  Self.Passengers.Remove(Passenger);
                  Inc(ServedPassengerCount);
                end;
              end;
            end;
          end;

          // Unload passengers to change line
          if Assigned(CurrentStation) then begin
            for P := Passengers.Count - 1 downto 0 do begin
              if not CurrentStation.MapStation.IsBestStationForShape(Passengers[P].Shape,
              TargetStation, CurrentStation) then begin
                Passenger := Passengers[P];
                Passengers.Delete(P);
                CurrentStation.MapStation.Passengers.Add(Passenger);
              end;
            end;
            for J := 0 to Carriages.Count - 1 do
            with Carriages[J] do begin
              for P := Passengers.Count - 1 downto 0 do begin
                if not CurrentStation.MapStation.IsBestStationForShape(Passengers[P].Shape,
                TargetStation, CurrentStation) then begin
                  Passenger := Passengers[P];
                  Passengers.Delete(P);
                  CurrentStation.MapStation.Passengers.Add(Passenger);
                end;
              end;
            end;
          end;

          // Load new passengers
          if Assigned(CurrentStation) and not Assigned(CurrentStation.MapStation) then
            raise Exception.Create(SStationWithoutMapStation);
          if Assigned(CurrentStation)  then
          for P := CurrentStation.MapStation.Passengers.Count - 1 downto 0 do begin
            if (Passengers.Count < TrainPassengerCount) then begin
              Passenger := CurrentStation.MapStation.Passengers[P];
              if CurrentStation.MapStation.IsBestStationForShape(Passenger.Shape,
              TargetStation, CurrentStation) then begin
                CurrentStation.MapStation.Passengers.Delete(P);
                Passengers.Add(Passenger);
              end;
            end else begin
              Done := False;
              for J := 0 to Carriages.Count - 1 do
              with Carriages[J] do begin
                if (Passengers.Count < TrainPassengerCount) then begin
                  Passenger := CurrentStation.MapStation.Passengers[P];
                  if CurrentStation.MapStation.IsBestStationForShape(Passenger.Shape,
                  TargetStation, CurrentStation) then begin
                    CurrentStation.MapStation.Passengers.Delete(P);
                    Passengers.Add(Passenger);
                    Done := True;
                    Break;
                  end;
                end;
              end;
              if not Done then Break;
            end;
          end;

          LastPosDelta := Abs(GetTargetStationDistance);
          InStation := False;
          LastTrainMoveTime := Time;
        end;
      end else begin
        PosChange := Direction + Trunc(Direction * TrainSpeed * (Time - LastTrainMoveTime));
        LastTrainMoveTime := Time;
        Redraw;
        if Assigned(TrackPosition.BaseTrackPoint) then
          TrackPosition.Move(PosChange);

        if Assigned(TargetStation) then begin
          PosDelta := Abs(GetTargetStationDistance);
          if PosDelta >= LastPosDelta then begin
            // We are getting far from station, stop at station
            TrackPosition.BaseTrackPoint := TargetStation.TrackPoint;
            TrackPosition.RelPos := 0;
            InStation := True;
            StationStopTime := Time;
            Redraw;
          end;
        end;
        LastPosDelta := PosDelta;
      end;
    end;
  end;
end;

function TEngine.GetUnusedLine: TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Lines.Count) and (Lines[I].Track.Points.Count > 0) do Inc(I);
  if I < Lines.Count then Result := Lines[I]
    else Result := nil;
end;

procedure TEngine.ShiftTrackPoints;
var
  I: Integer;
  J: Integer;
  //Link1, Link2: TPoint;
  NewPoint: TPoint;
  MetroLine: TMetroLine;
  TrackPoint: TTrackPoint;
  MapStation: TMapStation;
begin
  // NewGame all trackpoints position shift
  for MetroLine in Lines do
    for TrackPoint in MetroLine.Track.Points do
      TrackPoint.Position := TrackPoint.PositionDesigned;

  // Calculate new position shifts
  for MapStation in Stations do
    MapStation.ShiftTrackPoints;

  // Compute track points from track shift
  for MetroLine in Lines do
  with MetroLine do begin
    // Update start
    if Track.Points.Count > 1 then begin
      Track.Points[0].Position := Track.Points[0].PositionDesigned +
        Track.Points[0].LinkUp.Shift;
    end;

    for I := 1 to Track.Points.Count - 1 do
    with Track.Points[I] do
    if Assigned(Track.Points[I].LinkDown) and Assigned(Track.Points[I].LinkUp) then begin
      {
      Link1 := (Track.Points[I].PositionDesigned + Track.Points[I].LinkDown.Shift) -
        (Track.Points[I - 1].PositionDesigned + Track.Points[I].LinkDown.Shift);
      if (I + 1) < Track.Points.Count then
      Link2 := (Track.Points[I + 1].PositionDesigned + Track.Points[I].LinkUp.Shift) -
          (Track.Points[I].PositionDesigned + Track.Points[I].LinkUp.Shift)
        else Link2 := Link1;

      if ArcTanPoint(Link1) = ArcTanPoint(Link2) then begin
        // Parallel lines
        NewPoint := Track.Points[I].PositionDesigned + Track.Points[I].LinkDown.Shift;
        Track.Points[I].Position := NewPoint;
      end else begin}
        // Intersected lines
        if LineIntersect(Track.Points[I - 1].PositionDesigned + Track.Points[I].LinkDown.Shift,
          Track.Points[I].PositionDesigned + Track.Points[I].LinkDown.Shift,
          Track.Points[I].PositionDesigned + Track.Points[I].LinkUp.Shift,
          Track.Points[I + 1].PositionDesigned + Track.Points[I].LinkUp.Shift, NewPoint) then begin
            Track.Points[I].Position := NewPoint;
          end else begin
            // Parallel lines
            NewPoint := Track.Points[I].PositionDesigned + Track.Points[I].LinkDown.Shift;
            Track.Points[I].Position := NewPoint;
          end;
//      end;

      // Update ending
      if Track.Points.Count > 1 then begin
        Track.Points[Track.Points.Count - 1].Position := Track.Points[Track.Points.Count - 1].PositionDesigned -
          Track.Points[Track.Points.Count - 1].LinkDown.Shift;
      end;
    end;

    // Update ending
    if Track.Points.Count > 1 then begin
      Track.Points[Track.Points.Count - 1].Position := Track.Points[Track.Points.Count - 1].PositionDesigned -
        Track.Points[Track.Points.Count - 1].LinkDown.Shift;
    end;
  end;

  // Remove all temporal links
  for MetroLine in Lines do
  with MetroLine do begin
    for J := 0 to Track.Points.Count - 1 do
    if Assigned(Track.Points[J].LinkUp) then begin
      Track.Points[J].LinkUp.Free;
      Track.Points[J].LinkUp := nil;
      Track.Points[J + 1].LinkDown := nil;
    end;
  end;
end;

procedure TEngine.MenuItemExit(Sender: TObject);
begin
  if Assigned(FOnExit) then FOnExit(Self);
end;

procedure TEngine.MenuItemCity(Sender: TObject);
begin
  if Sender is TMenuItemButton then begin;
    City := TCity(TMenuItemButton(Sender).Ref);
    NewGame;
  end;
end;

procedure TEngine.MenuItemPlay(Sender: TObject);
begin
  City := nil;
  NewGame;
end;

procedure TEngine.MenuItemCareer(Sender: TObject);
var
  City: TCity;
begin
  MenuCareer.Parent := MenuMain;
  with MenuCareer, Items do begin
    Clear;
    for City in Cities do
    with AddButton(City.Name, MenuItemCity) do begin
      Ref := City;
      Enabled := not City.Locked;
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SBack, MenuItemBack) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    OnExit := MenuItemBack;
  end;

  Menu := MenuCareer;
  Redraw;
end;

procedure TEngine.MenuItemOptions(Sender: TObject);
begin
  MenuOptions.Parent := MenuMain;
  with MenuOptions, Items do begin
    Clear;
    with AddComboBox(SLanguage, [], LanguageChanged) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
      if Assigned(Translator) then begin
        Translator.LanguageListToStrings(States);
        Index := States.IndexOfObject(Translator.Language);
      end;
      if Index = -1 then Index := 0;
    end;
    with AddCheckBox(SDarkMode, DarkModeChanged) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
      Checked := DarkMode;
    end;
    with AddCheckBox(SFullScreen, FullScreenChanged) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
      Checked := FullScreen;
    end;
    with AddButton(SBack, MenuItemBack) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    OnExit := MenuItemBack;
  end;

  Menu := MenuOptions;
  Redraw;
end;

procedure TEngine.MenuItemBack(Sender: TObject);
begin
  if Assigned(Menu.Parent) then begin
    Menu := Menu.Parent;
    Redraw;
  end else MenuItemExit(nil);
end;

procedure TEngine.ButtonPlay(Sender: TObject);
begin
  TimePerSecond := TimePerSecondNormal;
  if State = gsPaused then State := gsRunning;
  UpdateInterface;
end;

procedure TEngine.ButtonPause(Sender: TObject);
begin
  if State = gsRunning then State := gsPaused;
  UpdateInterface;
end;

procedure TEngine.ButtonFastForward(Sender: TObject);
begin
  TimePerSecond := TimePerSecondFast;
  if State = gsPaused then State := gsRunning;
  UpdateInterface;
end;

procedure TEngine.ButtonNewTrain(Sender: TObject);
var
  Improvements: TList<TMetroImprovement>;
  Improvement: TMetroImprovement;
  Index: Integer;
begin
  Trains.AddNew;
  State := gsNewImprovement;
  if Lines.Count <= (High(LineColors) - Low(LineColors)) then
    AvailableImprovements := AvailableImprovements + [miLine]
    else AvailableImprovements := AvailableImprovements - [miLine];

  // Select offered improvements
  Improvements := TList<TMetroImprovement>.Create;
  try
    for Improvement := Low(TMetroImprovement) to High(TMetroImprovement) do
    if Improvement in AvailableImprovements then begin
      Improvements.Add(Improvement);
    end;

    Index := Random(Integer(Improvements.Count));
    Improvement1 := Improvements[Index];
    Improvements.Delete(Index);

    if Improvements.Count > 0 then begin
      Index := Random(Integer(Improvements.Count));
      Improvement2 := Improvements[Index];
    end else Improvement2 := miNone;

    if Improvement1 <> miNone then
      ImageNewImprovement1.Bitmap.Assign(GetImprovementBitmap(Improvement1));
    ImageNewImprovement1.Enabled := Improvement1 <> miNone;

    if Improvement2 <> miNone then
      ImageNewImprovement2.Bitmap.Assign(GetImprovementBitmap(Improvement2));
    ImageNewImprovement2.Enabled := Improvement2 <> miNone;
  finally
    Improvements.Free;
  end;
  Redraw;
end;

procedure TEngine.ButtonNewImprovement1(Sender: TObject);
begin
  EvaluateImprovement(Improvement1);
  State := gsRunning;
  Redraw;
end;

procedure TEngine.ButtonNewImprovement2(Sender: TObject);
begin
  EvaluateImprovement(Improvement2);
  State := gsRunning;
  Redraw;
end;

procedure TEngine.MenuItemGameContinue(Sender: TObject);
begin
  State := LastState;
  Redraw;
end;

procedure TEngine.MenuItemGameExit(Sender: TObject);
begin
  State := gsMenu;
  Clear;
  Menu := MenuMain;
  Redraw;
end;

procedure TEngine.MenuItemGameRestart(Sender: TObject);
begin
  NewGame;
end;

procedure TEngine.DarkModeChanged(Sender: TObject);
begin
  DarkMode := TMenuItemCheckBox(Sender).Checked;
  InitMenus;
end;

procedure TEngine.LanguageChanged(Sender: TObject);
var
  NewLanguage: TLanguage;
begin
  NewLanguage := TLanguage(TMenuItemComboBox(Sender).States.Objects[TMenuItemComboBox(Sender).Index]);
  if Assigned(Translator) and (Translator.Language <> NewLanguage) then begin
    Translator.Language := NewLanguage;
    Translator.Translate;
    InitMenus;

    // Recreate cities with translated names
    SaveToRegistry;
    InitCities;
    LoadFromRegistry;
  end;
end;

procedure TEngine.FullScreenChanged(Sender: TObject);
begin
  FullScreen := TMenuItemCheckBox(Sender).Checked;
end;

procedure TEngine.UpdateInterface;
begin
  ImagePlay.Enabled := not ((State = gsRunning) and (TimePerSecond = TimePerSecondNormal));
  ImageFastForward.Enabled := not ((State = gsRunning) and (TimePerSecond = TimePerSecondFast));
  ImagePause.Enabled := FState = gsRunning; //not (State = gsPaused);
end;

procedure TEngine.InitCities;
var
  I: Integer;
begin
  with Cities do begin
    Clear;
    with AddNew('Prague', SPrague, 1275406) do begin
      LineColors := [clRed, clGreen, clYellow];
      InitialLineCount := 1;
      PassengersCountToUnlock := 300;
    end;
    with AddNew('Paris', SParis, 2138551) do begin
      LineColors := [clRed, clGreen, clYellow, clBlue, clBrown, clOrange,
        clPurple, clOlive, clAqua, clDarkYellow, clPink];
      InitialLineCount := 1;
      PassengersCountToUnlock := 500;
    end;
    with AddNew('Rome', SRome, 2872800) do begin
      LineColors := [clRed, clGreen, clYellow];
      InitialLineCount := 1;
      PassengersCountToUnlock := 700;
    end;
    with AddNew('NewYork', SNewYork, 8804190) do begin
      InitialLineCount := 1;
      PassengersCountToUnlock := 1000;
    end;
    with AddNew('London', SLondon, 9002488) do begin
      InitialLineCount := 3;
      PassengersCountToUnlock := 1500;
    end;
    with AddNew('Seoul', SSeoul, 10197604) do begin
      InitialLineCount := 3;
      LineColors := [clRed, clGreen, clYellow, clOrange, clGray,
        clPink, clCyan, clMoneyGreen, clPurple, clSkyBlue];
      PassengersCountToUnlock := 2000;
    end;
    with AddNew('Tokyo', STokyo, 13960236) do begin
      InitialLineCount := 3;
      PassengersCountToUnlock := 2500;
    end;
    with AddNew('Beijing', SBeijing, 21893095) do begin
      InitialLineCount := 3;
      LineColors := [clRed, clGreen, clYellow, clOrange, clOlive, clGray,
        clPink, clCyan, clMoneyGreen, clPurple, clSkyBlue, clBrown];
      PassengersCountToUnlock := 3000;
    end;
  end;
  for I := 1 to Cities.Count - 1 do
    Cities[I].Locked := True;
end;

function TEngine.GetImprovementText(Improvement: TMetroImprovement): string;
begin
  case Improvement of
    miCarriage: Result := SCarriage;
    miTerminal: Result := STerminal;
    miTunnel: Result := STunnel;
    miLine: Result := SLine;
  end;
end;

function TEngine.GetImprovementBitmap(Improvement: TMetroImprovement): TBitmap;
begin
  case Improvement of
    miCarriage: Result := ImageCarriage.Bitmap;
    miTerminal: Result := ImageTerminal.Bitmap;
    miTunnel: Result := ImageTunnel.Bitmap;
    miLine: Result := ImageLine.Bitmap;
  end;
end;

procedure TEngine.EvaluateImprovement(Improvement: TMetroImprovement);
begin
  case Improvement of
    miLine: Lines.AddNew(LineColors[Lines.Count]);
    miCarriage: Carriages.AddNew;
    //miTunnel: Tunnels.AddNew;
    miTerminal: Inc(AvailableTerminals);
  end;
end;

function TEngine.GetSelectedOrUnusedMetroLine: TMetroLine;
begin
  if Assigned(SelectedLine) and (SelectedLine.LineStations.Count = 0) then
    Result := SelectedLine
    else Result := GetUnusedLine;
end;

procedure TEngine.InitMenus;
begin
  with MenuMain, Items do begin
    Clear;
    with AddButton(SBigMetro, nil) do begin
      Enabled := False;
      TextSize := 60;
      TextColor := Colors.Text;
      TextDisabledColor := Colors.Text;
      BackgroundColor := clNone;
      BackgroundSelectedColor := clNone;
    end;
    with AddButton(SPlay, MenuItemPlay) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SCareer, MenuItemCareer) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SOptions, MenuItemOptions) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SExit, MenuItemExit) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    OnExit := MenuItemExit;
  end;

  with MenuGame, Items do begin
    Clear;
    with AddButton(SContinue, MenuItemGameContinue) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SRestart, MenuItemGameRestart) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    with AddButton(SExit, MenuItemGameExit) do begin
      TextSize := 40;
      TextColor := Colors.MenuItemText;
      TextDisabledColor := Colors.MenuItemDisabledText;
      BackgroundColor := Colors.MenuItemBackground;
      BackgroundSelectedColor := Colors.MenuItemBackgroundSelected;
    end;
    OnExit := MenuItemGameContinue;
  end;
end;

procedure TEngine.ButtonBackClick(Sender: TObject);
begin
  if State = gsSuccess then begin
    State := gsMenu;
    MenuItemCareer(Self);
  end else begin
    Menu := MenuGame;
    LastState := State;
    State := gsMenu;
    Redraw;
  end;
end;

procedure TEngine.DrawClock(Canvas: TCanvas; CanvasSize: TPoint);
var
  ClockCenter: TPoint;
  Angle: Double;
  Text: string;
  I: Integer;
  Y: Integer;
  IsDay: Boolean;
const
  ClockSize = 20;
begin
  IsDay := (((Time / OneHour) mod 24) > 6) and (((Time / OneHour) mod 24) < 18);
  if IsDay then begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := Colors.Background;
  end else begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := Colors.Text;
  end;
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Color := Colors.Text;
  Canvas.Pen.Width := 2;
  ClockCenter := Point(CanvasSize.X - 30, 40);
  Angle := Time / (12 * OneHour) * 2 * Pi - Pi / 2;
  Canvas.EllipseC(ClockCenter.X, ClockCenter.Y, ClockSize, ClockSize);
  Canvas.Brush.Style := bsClear;

  if IsDay then begin
    Canvas.Pen.Color := Colors.Text;
  end else begin
    Canvas.Pen.Color := Colors.Background;
  end;
  Canvas.Line(ClockCenter, Point(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.8),
    ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.8)));
  for I := 0 to 12 do begin
    Angle := I / 12 * 2 * Pi;
    Canvas.Line(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.7),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.7),
      ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.9),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.9));
  end;
  Y := ClockCenter.Y;

  Canvas.Pen.Color := Colors.Text;
  Canvas.Font.Color := Colors.Text;
  Text := FormatDateTime('ddd', Time + 2);
  Canvas.TextOut(ClockCenter.X - ClockSize - Canvas.TextWidth(Text) - ScaleX(10, 96),
    Y - Canvas.TextHeight(Text) div 2, Text);
  Y := Y + Canvas.TextHeight(Text) + ScaleY(5, 96);

  Text := SDay + ' ' + IntToStr(Trunc(Time));
  Canvas.TextOut(ClockCenter.X - ClockSize - Canvas.TextWidth(Text) - ScaleX(10, 96),
    Y - Canvas.TextHeight(Text) div 2, Text);
  Y := Y + Canvas.TextHeight(Text) + ScaleY(5, 96);

  ImagePause.Canvas := Canvas;
  ImagePause.Bounds := Bounds(CanvasSize.X - ScaleX(20 + 10, 96), Y,
    ScaleX(20, 96), ScaleY(20, 96));
  ImagePause.Paint;
  Y := Y + ImagePause.Bounds.Height + ScaleY(5, 96);

  ImagePlay.Canvas := Canvas;
  ImagePlay.Bounds := Bounds(CanvasSize.X - ScaleX(20 + 10, 96) , Y,
    ScaleX(20, 96), ScaleY(20, 96));
  ImagePlay.Paint;
  Y := Y + ImagePlay.Bounds.Height + ScaleY(5, 96);

  ImageFastForward.Canvas := Canvas;
  ImageFastForward.Bounds := Bounds(CanvasSize.X - ScaleX(20 + 10, 96) , Y,
    ScaleX(20, 96), ScaleY(20, 96));
  ImageFastForward.Paint;
  Y := Y + ImageFastForward.Bounds.Height + ScaleY(5, 96);
end;

procedure TEngine.DrawTrains(Canvas: TCanvas);
var
  P: Integer;
  Pos: TPoint;
  Points: array of TPoint;
  Angle: Double;
  ShapePos: TPoint;
  Train: TMetroTrain;
  Passenger: TMetroPassenger;
  Carriage: TMetroCarriage;
  Vector: TVector;
begin
  for Train in Trains do
  with Train do begin
    if Assigned(Line) then begin
      Canvas.Brush.Color := Line.Color;
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Style := psClear;
      Vector := TrackPosition.GetVector;
      Pos := Vector.Position;
      Angle := Vector.GetAngle;

      SetLength(Points, 4);
      Points[0] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
      Points[1] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
      Points[2] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
      Points[3] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
      Canvas.Polygon(Points);
      Canvas.Brush.Color := clWhite;
      P := 0;
      for Passenger in Passengers do
      with Passenger do begin
        ShapePos := Point(Pos.X - Trunc(TrainSize div 3 * 1) + (P mod 3) * TrainSize div 3,
          Pos.Y - Trunc(TrainSize div 6 * 1) + (P div 3) * TrainSize div 3);
        ShapePos := RotatePoint(Pos, ShapePos, Angle);
        DrawShape(Canvas, ShapePos, Shape, TrainSize div 3, Angle + Pi / 2);
        Inc(P);
      end;

      // Draw carriages
      for Carriage in Train.Carriages do
      with Carriage do begin
        Canvas.Brush.Color := Line.Color;
        Canvas.Brush.Style := bsSolid;
        Canvas.Pen.Style := psClear;
        Vector := GetTrackPosition.GetVector;
        Pos := Vector.Position;
        Angle := Vector.GetAngle;

        SetLength(Points, 4);
        Points[0] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
        Points[1] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
        Points[2] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
        Points[3] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
        Canvas.Polygon(Points);
        Canvas.Brush.Color := clWhite;
        P := 0;
        for Passenger in Passengers do
        with Passenger do begin
          ShapePos := Point(Pos.X - Trunc(TrainSize div 3 * 1) + (P mod 3) * TrainSize div 3,
            Pos.Y - Trunc(TrainSize div 6 * 1) + (P div 3) * TrainSize div 3);
          ShapePos := RotatePoint(Pos, ShapePos, Angle);
          DrawShape(Canvas, ShapePos, Shape, TrainSize div 3, Angle + Pi / 2);
          Inc(P);
        end;
      end;

{      // Target station links
      if Assigned(TargetStation) then begin
        Canvas.Pen.Width := 1;
        Canvas.Pen.Style := psSolid;
        Canvas.Pen.Color := Colors.MenuItemText;
        Canvas.MoveTo(Pos);
        Canvas.LineTo(TargetStation.TrackPoint.Position);
      end;
}
    end;
  end;
end;

procedure TEngine.DrawGameOver(Canvas: TCanvas; CanvasSize: TPoint);
var
  Y: Integer;
  Text: string;
begin
  with Canvas do begin
    DrawFrame(Canvas, Bounds(CanvasSize.X div 8, ScaleY(90, 96),
      Round(CanvasSize.X / 4 * 3), ScaleY(210, 96)));

    Canvas.Font.Color := Self.Colors.Text;
    Brush.Style := bsClear;
    Pen.Style := psClear;

    Y := ScaleY(100, 96);

    Font.Size := 40;
    Text := SGameOver;
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(Text) * 1.1);

    Font.Size := 14;
    Text := SGameOverReason;
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(Text) * 1.1);

    Text := Format(SGameOverStatistic, [ServedPassengerCount, ServedDaysCount]);
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(SGameOverStatistic) * 1.1);

    Y := Y + ScaleY(16, 96);

    DrawHighScore(Canvas, CanvasSize, Y);
  end;
end;

procedure TEngine.DrawHighScore(Canvas: TCanvas; CanvasSize: TPoint; Y: Integer);
var
  Text: string;
begin
  with Canvas do begin
    // Calculate new highest score
    Text := '';
    if (ServedPassengerCount > HighestServedPassengerCount) or
      (ServedDaysCount > HighestServedDaysCount) then begin
      Text := SNewHighScore + ' ';
    end;
    Text := Text + Format(SOldHighScore, [HighestServedPassengerCount,
      HighestServedDaysCount]);
    Canvas.TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(Text) * 1.1);
    if (ServedPassengerCount > HighestServedPassengerCount) then
      HighestServedPassengerCount := ServedPassengerCount;
    if (ServedDaysCount > HighestServedDaysCount) then
      HighestServedDaysCount := ServedDaysCount;

    if Assigned(City) then begin
      if (ServedPassengerCount > City.HighestServedPassengerCount) then
        City.HighestServedPassengerCount := ServedPassengerCount;
      if (ServedDaysCount > City.HighestServedDaysCount) then
        City.HighestServedDaysCount := ServedDaysCount;
    end;
  end;
end;

procedure TEngine.DrawSuccess(Canvas: TCanvas; CanvasSize: TPoint);
var
  Y: Integer;
  Text: string;
  Index: Integer;
begin
  with Canvas do begin
    DrawFrame(Canvas, Bounds(CanvasSize.X div 8, ScaleY(90, 96),
      Round(CanvasSize.X / 4 * 3), ScaleY(230, 96)));

    Canvas.Font.Color := Self.Colors.Text;
    Brush.Style := bsClear;
    Pen.Style := psClear;

    Y := ScaleY(100, 96);

    Font.Size := 40;
    Text := SSuccess;
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(Text) * 1.1);

    Font.Size := 14;
    Text := SSuccessReason;
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(Text) * 1.1);

    // Unlock next city
    if Assigned(City) then begin
      Index := Cities.IndexOf(City);
      if (Index >= 0) and ((Index + 1) < Cities.Count) then begin
        Cities[Index + 1].Locked := False;
        Font.Size := 14;
        Text := Format(SUnlockedCity, [Cities[Index + 1].Name]);
        TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
        Y := Y + Round(TextHeight(Text) * 1.1);
      end;
    end;

    Text := Format(SGameOverStatistic, [ServedPassengerCount, ServedDaysCount]);
    TextOut((CanvasSize.X - TextWidth(Text)) div 2, Y, Text);
    Y := Y + Round(TextHeight(SGameOverStatistic) * 1.1);

    Y := Y + ScaleY(16, 96);

    DrawHighScore(Canvas, CanvasSize, Y);
  end;
end;

procedure TEngine.DrawNewWeek(Canvas: TCanvas; CanvasSize: TPoint);
var
  Text: string;
  TextSize: TSize;
  Y: Integer;
  X: Integer;
begin
  with Canvas do begin

    DrawFrame(Canvas, Bounds(CanvasSize.X div 4, CanvasSize.Y div 4, CanvasSize.X div 2, CanvasSize.Y div 2));
    X := CanvasSize.X div 4 + ScaleX(10, 96);
    Y := CanvasSize.Y div 4 + ScaleX(10, 96);

    Text := IntToStr(Week) + '. ' + SWeek;
    Font.Size := 30;
    Font.Color := Self.Colors.Text;
    TextSize := TextExtent(Text);
    TextOut(X, Y, Text);
    Y := Y + Round(1.1 * TextSize.Height);

    Text := SNewTrain;
    Font.Size := 14;
    Font.Color := Self.Colors.Text;
    TextSize := TextExtent(Text);
    TextOut(X, Y, Text);
    Y := Y + Round(1.1 * TextSize.Height);

    ImageNewTrain.Canvas := Canvas;
    ImageNewTrain.Bounds := Bounds(CanvasSize.X div 2 - ImprovementImageSize div 2,
      CanvasSize.Y div 2 - ImprovementImageSize div 2,
      ImprovementImageSize, ImprovementImageSize);
    ImageNewTrain.Paint;

    Y := (CanvasSize.Y - ImprovementImageSize) div 2 +
      Round(ImprovementImageSize * 1.1);

    Text := STrain;
    Font.Size := 20;
    Font.Color := Self.Colors.Text;
    TextSize := TextExtent(Text);
    TextOut((CanvasSize.X - TextSize.Width) div 2, Y, Text);
  end;
end;

procedure TEngine.DrawNewImprovement(Canvas: TCanvas; CanvasSize: TPoint);
var
  Text: string;
  TextSize: TSize;
  Y: Integer;
  X: Integer;
  Center: TPoint;
begin
  with Canvas do begin
    Center := Point(CanvasSize.X div 2, CanvasSize.Y div 2);

    DrawFrame(Canvas, Bounds(CanvasSize.X div 4, CanvasSize.Y div 4, CanvasSize.X div 2, CanvasSize.Y div 2));
    X := CanvasSize.X div 4 + ScaleX(10, 96);
    Y := CanvasSize.Y div 4 + ScaleX(10, 96);

    Text := IntToStr(Week) + '. ' + SWeek;
    Font.Size := 30;
    Font.Color := Self.Colors.Text;
    TextSize := TextExtent(Text);
    TextOut(X, Y, Text);
    Y := Y + Round(1.1 * TextSize.Height);

    Text := SNewImprovement;
    Font.Size := 14;
    Font.Color := Self.Colors.Text;
    TextSize := TextExtent(Text);
    TextOut(X, Y, Text);
    Y := Y + Round(1.1 * TextSize.Height);

    if Improvement1 <> miNone then begin
      ImageNewImprovement1.Canvas := Canvas;
      ImageNewImprovement1.Bounds := Bounds(Center.X - CanvasSize.X div 8 -
        ImprovementImageSize div 2, Center.Y - ImprovementImageSize div 2,
        ImprovementImageSize, ImprovementImageSize);
      ImageNewImprovement1.Paint;
    end;

    if Improvement2 <> miNone then begin
      ImageNewImprovement2.Canvas := Canvas;
      ImageNewImprovement2.Bounds := Bounds(Center.X + CanvasSize.X div 8 -
        ImprovementImageSize div 2, Center.Y - ImprovementImageSize div 2,
        ImprovementImageSize, ImprovementImageSize);
     ImageNewImprovement2.Paint;
    end;

    Y := Center.Y - ImprovementImageSize div 2 +
      Round(ImprovementImageSize * 1.1);

    if Improvement1 <> miNone then begin
      Text := GetImprovementText(Improvement1);
      Font.Size := 20;
      Font.Color := Self.Colors.Text;
      TextSize := TextExtent(Text);
      TextOut(Center.X - CanvasSize.X div 8 - TextSize.Width div 2, Y, Text);
    end;

    if Improvement2 <> miNone then begin
      Text := GetImprovementText(Improvement2);
      Font.Size := 20;
      Font.Color := Self.Colors.Text;
      TextSize := TextExtent(Text);
      TextOut(Center.X + CanvasSize.X div 8 - TextSize.Width div 2, Y, Text);
    end;
  end;
end;

procedure TEngine.DrawStationPassengerOverload(Canvas: TCanvas);
var
  MapStation: TMapStation;
  Angle: Real;
begin
  for MapStation in Stations do
  with MapStation do begin
    if OverloadDuration > 0 then begin
      Canvas.Brush.Color := clSilver;
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Color := clSilver;
      Canvas.Pen.Style := psSolid;
      Angle := OverloadDuration / MaxPassengersOveloadTime * 2 * Pi;
      Canvas.Pie(Position.X - StationOverloadSize, Position.Y - StationOverloadSize,
      Position.X + StationOverloadSize, Position.Y + StationOverloadSize,
        Trunc(Position.X + StationOverloadSize * Cos(Angle)),
        Trunc(Position.Y + StationOverloadSize * Sin(Angle)), Position.X + StationOverloadSize, Position.Y);
    end;
  end;
end;

procedure TEngine.DrawLines(Canvas: TCanvas);
var
  MetroLine: TMetroLine;
  S: Integer;
begin
  for MetroLine in Lines do
  with MetroLine do begin
    Canvas.Pen.Color := Color;
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := MetroLineThickness;
    if Track.Points.Count > 0 then Canvas.MoveTo(Track.Points[0].Position);
    for S := 1 to Track.Points.Count - 1 do begin
      Canvas.LineTo(Track.Points[S].Position);
{      if (S = TrackPoints.Count - 1) then begin
        Canvas.Pen.EndCap := pecSquare;
        Angle := arctan2D(((TrackPoints[S].Position.Y - TrackPoints[S - 1].Position.Y),
          (TrackPoints[S].Position.X - TrackPoints[S - 1].Position.X));
        EndPoint := Point(Round(TrackPoints[S].Position.X + EndStationLength * Cos(Angle)),
          Round(TrackPoints[S].Position.Y + EndStationLength * Sin(Angle)));
        Canvas.LineTo(EndPoint);
        Canvas.MoveTo(Point(Round(EndPoint.X + Cos(Angle + Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle + Pi / 2) * EndStationLength / 3)));
        Canvas.LineTo(Point(Round(EndPoint.X + Cos(Angle - Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle - Pi / 2) * EndStationLength / 3)));
        Canvas.Pen.EndCap := pecRound;
      end;}
    end;
(*    Canvas.Pen.Color := Color;
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := MetroLineThickness div 2;
    if Track.Points.Count > 0 then Canvas.MoveTo(Track.Points[0].PositionDesigned);
    for S := 1 to Track.Points.Count - 1 do begin
      Canvas.LineTo(Track.Points[S].PositionDesigned);
{      if (S = TrackPoints.Count - 1) then begin
        Canvas.Pen.EndCap := pecSquare;
        Angle := arctan2((TrackPoints[S].Position.Y - TrackPoints[S - 1].Position.Y),
          (TrackPoints[S].Position.X - TrackPoints[S - 1].Position.X));
        EndPoint := Point(Round(TrackPoints[S].Position.X + EndStationLength * Cos(Angle)),
          Round(TrackPoints[S].Position.Y + EndStationLength * Sin(Angle)));
        Canvas.LineTo(EndPoint);
        Canvas.MoveTo(Point(Round(EndPoint.X + Cos(Angle + Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle + Pi / 2) * EndStationLength / 3)));
        Canvas.LineTo(Point(Round(EndPoint.X + Cos(Angle - Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle - Pi / 2) * EndStationLength / 3)));
        Canvas.Pen.EndCap := pecRound;
      end;}
    end;
    {
    if (TrackPoints.Count > 1) then begin
      Canvas.Pen.EndCap := pecSquare;
      Angle := arctan2((TrackPoints[1].Position.Y - TrackPoints[0].Position.Y),
        (TrackPoints[1].Position.X - TrackPoints[0].Position.X));
      Canvas.MoveTo(TrackPoints[0].Position);
      EndPoint := Point(Round(TrackPoints[0].Position.X - EndStationLength * Cos(Angle)),
        Round(TrackPoints[0].Position.Y - EndStationLength * Sin(Angle)));
      Canvas.LineTo(EndPoint);
      Canvas.MoveTo(Point(Round(EndPoint.X - Cos(Angle + Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle + Pi / 2) * EndStationLength / 3)));
      Canvas.LineTo(Point(Round(EndPoint.X - Cos(Angle - Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle - Pi / 2) * EndStationLength / 3)));
      Canvas.Pen.EndCap := pecRound;
    end;  }
    *)
  end;

  // Draw design time lines
  if Assigned(TrackStationDown) and Assigned(TrackStationDown.OwnerPoint) then begin
    Canvas.Pen.Color := TMetroLine(TrackStationDown.Track.Owner).Color;
    Canvas.MoveTo(TLineStation(TrackStationDown.OwnerPoint).TrackPoint.Position);
    DrawLine(Canvas, View.PointDestToSrc(LastMousePos));
  end;
  if Assigned(TrackStationUp) and Assigned(TrackStationUp.OwnerPoint) then begin
    Canvas.Pen.Color := TMetroLine(TrackStationUp.Track.Owner).Color;
    Canvas.MoveTo(TLineStation(TrackStationUp.OwnerPoint).TrackPoint.Position);
    DrawLine(Canvas, View.PointDestToSrc(LastMousePos));
  end;
end;

procedure TEngine.DrawStations(Canvas: TCanvas);
var
  MapStation: TMapStation;
  Passenger: TMetroPassenger;
  PassengerPos: TPoint;
  Direction: Integer;
  UsedStationSize: Integer;
begin
  Canvas.Pen.Width := 5;
  for MapStation in Stations do
  with MapStation do begin
    Canvas.Pen.Style := psSolid;
    if IsTerminal then UsedStationSize := Round(StationSize * 1.5)
      else UsedStationSize := StationSize;

    if Assigned(SelectedLine) and (Lines.IndexOf(SelectedLine) <> -1) then begin
      Canvas.Brush.Style := bsClear;
      Canvas.Pen.Color := SelectedLine.Color;
      DrawShape(Canvas, Position, Shape, UsedStationSize + Canvas.Pen.Width + 4, 0);
    end;

    Canvas.Brush.Color := Colors.ShapeBackground;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Color := Colors.Text;
    DrawShape(Canvas, Position, Shape, UsedStationSize, 0);

    // Draw passengers
    Canvas.Pen.Style := psClear;
    Canvas.Brush.Color := Colors.Text;
    PassengerPos := Point(0, 0);
    Direction := 1;
    for Passenger in Passengers do
    with Passenger do begin
      DrawShape(Canvas, Point(Position.X + StationSize + PassengerPos.X,
        Position.Y - StationSize div 2 + PassengerPos.Y),
        Shape, PassengerSize, 0);
      PassengerPos := Point(PassengerPos.X + Direction * (PassengerSize + 2), PassengerPos.Y);
      if PassengerPos.X >= (PassengerSize + 2) * VisiblePassengersPerLine then begin
        Direction := -Direction;
        PassengerPos.X := PassengerPos.X - (PassengerSize + 2);
        PassengerPos.Y := PassengerPos.Y + (PassengerSize + 2);
      end;
      if PassengerPos.X < 0 then begin
        Direction := -Direction;
        PassengerPos.X := 0;
        PassengerPos.Y := PassengerPos.Y + (PassengerSize + 2);
      end;
    end;

{    if ShowDistances then begin
      Canvas.Brush.Style := bsClear;
      Text := '';
      for P := 0 to 5 do
        Text := Text + IntToStr(ShapeDistance[TStationShape(P)]) + ',';
      Canvas.TextOut(Position.X + StationSize div 2, Position.Y + StationSize div 2, Text);
    end;
    }
  end;
end;

procedure TEngine.DrawGameControls(Canvas: TCanvas; CanvasSize: TPoint);
var
  I: Integer;
  Text: string;
  Radius: Integer;
  Angle: Real;
  Pos: TPoint;
  X: Integer;
  Y: Integer;
  SeparatorSize: Integer;
begin
  SeparatorSize := ScaleX(20, 96);
  X := CanvasSize.X div 2;
  Y := CanvasSize.Y - LineColorsDist;

  // Line selection
  Canvas.Pen.Width := 4;
  for I := 0 to High(LineColors) do begin
    if Assigned(Lines.SearchByColor(LineColors[I])) then begin
      Canvas.Brush.Color := LineColors[I];
      Radius := 15;
    end else begin
      Canvas.Brush.Color := clSilver;
      Radius := 5;
    end;
    Canvas.Pen.Color := Colors.Text;
    if Assigned(SelectedLine) and (SelectedLine.Color = LineColors[I]) then begin
      Canvas.Pen.Style := psSolid;
    end else begin
      Canvas.Pen.Style := psClear;
    end;

    Canvas.EllipseC(X - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
      Y, Radius, Radius);
  end;
  X := X - Length(LineColors) div 2 * LineColorsDist - 2 * SeparatorSize;

  // Draw unused trains
  ImageLocomotive.Bounds := Bounds(X - IconSize, Y - IconSize div 2,
    IconSize, IconSize);
  ImageLocomotive.Canvas := Canvas;
  ImageLocomotive.Paint;
  X := X - IconSize - SeparatorSize div 3;

  Text := IntToStr(Trains.GetUnusedCount);
  Canvas.Brush.Style := bsClear;
  Canvas.Font.Size := 14;
  Canvas.Font.Color := Colors.Text;
  Canvas.TextOut(X - Canvas.TextWidth(Text),
    Y - Canvas.TextHeight(Text) div 2, Text);
  X := X - Canvas.TextWidth(Text) - SeparatorSize;

  // Draw unused carriages
  if Carriages.GetUnusedCount > 0 then CarriageCountVisible := True;
  if CarriageCountVisible then begin
    Text := IntToStr(Carriages.GetUnusedCount);
    ImageCarriage.Bounds := Bounds(X - IconSize, Y - IconSize div 2,
      IconSize, IconSize);
    ImageCarriage.Canvas := Canvas;
    ImageCarriage.Paint;
    X := X - IconSize - SeparatorSize div 3;

    Canvas.Brush.Style := bsClear;
    Canvas.Font.Size := 14;
    Canvas.Font.Color := Colors.Text;
    Canvas.TextOut(X - Canvas.TextWidth(Text),
      Y - Canvas.TextHeight(Text) div 2, Text);
    X := X - Canvas.TextWidth(Text) - SeparatorSize;
  end;

  // Draw unused terminals
  if AvailableTerminals > 0 then TerminalCountVisible := True;
  if TerminalCountVisible then begin
    Text := IntToStr(AvailableTerminals);
    ImageTerminal.Bounds := Bounds(X - IconSize, Y - IconSize div 2,
      IconSize, IconSize);
    ImageTerminal.Canvas := Canvas;
    ImageTerminal.Paint;
    X := X - IconSize - SeparatorSize div 3;

    Canvas.Brush.Style := bsClear;
    Canvas.Font.Size := 14;
    Canvas.Font.Color := Colors.Text;
    Canvas.TextOut(X - Canvas.TextWidth(Text),
      Y - Canvas.TextHeight(Text) div 2, Text);
  end;

  // Passenger count
  X := CanvasSize.X - ScaleX(10, 96);
  ImagePassenger.Bounds := Bounds(X - IconSize, Y - IconSize div 2,
    IconSize, IconSize);
  ImagePassenger.Canvas := Canvas;
  ImagePassenger.Paint;
  X := X - IconSize - SeparatorSize div 3;

  Canvas.Brush.Style := bsClear;
  Canvas.Font.Size := 14;
  Canvas.Font.Color := Colors.Text;
  Text := IntToStr(ServedPassengerCount);
  Canvas.TextOut(X - Canvas.TextWidth(Text),
    Y - Canvas.TextHeight(Text) div 2, Text);

  DrawClock(Canvas, CanvasSize);

  // Back button
  Canvas.Font.Size := 40;
  Canvas.Font.Color := Colors.Text;
  ButtonBack.Canvas := Canvas;
  ButtonBack.Bounds.Left := 10;
  ButtonBack.Bounds.Top := 10;
  ButtonBack.Paint;

  // City name
  if Assigned(City) then begin
    Canvas.Brush.Style := bsClear;
    Canvas.Font.Color := Colors.Text;
    Canvas.Font.Size := 20;
    Text := City.Name;
    Canvas.TextOut(20, CanvasSize.Y -
      Canvas.TextHeight(Text) - 20, Text);

    X := CanvasSize.X - ScaleX(10, 96);
    ImageAchievement.Bounds := Bounds(X - IconSize, Y - 2 * IconSize,
      IconSize, IconSize);
    ImageAchievement.Canvas := Canvas;
    ImageAchievement.Paint;

    X := X - IconSize - SeparatorSize div 3;

    Canvas.Brush.Style := bsClear;
    Canvas.Font.Size := 14;
    Canvas.Font.Color := Colors.Text;
    Text := IntToStr(City.PassengersCountToUnlock);
    Canvas.TextOut(X - Canvas.TextWidth(Text),
      Y - Round(1.5 * IconSize) - Canvas.TextHeight(Text) div 2, Text);
  end;
end;

procedure TEngine.DrawGrabbed(Canvas: TCanvas; CanvasSize: TPoint);
var
  Pos: TPoint;
  Angle: Double;
  FocusedTrack: TTrackLink;
  Vector: TVector;
  Intersect: TPoint;
  TrackPosition: TTrackPosition;
begin
  // Show train grabbed by mouse
  Angle := 0;
  Pos := LastMousePos;

  if Assigned(SelectedTrain) then begin
    FocusedTrack := GetTrackOnPos(View.PointDestToSrc(Pos), Intersect);
    if Assigned(FocusedTrack.Points[0]) then begin
      TrackPosition.BaseTrackPoint := FocusedTrack.Points[0];
      TrackPosition.RelPos := Distance(FocusedTrack.Points[0].Position,
        Intersect);
      Vector := TrackPosition.GetVector;
      Angle := Vector.GetAngle;

      if Assigned(LastGrabbedTrain) then begin
        if TrackPosition.GetTrackPosition >
          LastGrabbedTrainTrackPosition.GetTrackPosition then
            GrabbedTrainDirection := 1
        else
        if TrackPosition.GetTrackPosition <
          LastGrabbedTrainTrackPosition.GetTrackPosition then
          GrabbedTrainDirection := -1;
      end else GrabbedTrainDirection := 1;

      LastGrabbedTrain := SelectedTrain;
      LastGrabbedTrainTrackPosition := TrackPosition;
    end;
    FocusedTrack.Free;

    if GrabbedTrainDirection = -1 then Angle := Angle + Pi;
    Canvas.Brush.Color := Colors.Text; //SelectedTrain.Line.Color;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Style := psClear;
    Canvas.Polygon([
      RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X + TrainSize div 2 - 10, Pos.Y - TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y), Angle),
      RotatePoint(Pos, Point(Pos.X + TrainSize div 2 - 10, Pos.Y + TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y + TrainSize div 3), Angle)
    ]);
  end;

  // Show carriage grabbed by mouse
  if Assigned(SelectedCarriage) then begin
    Canvas.Brush.Color := Colors.Text; //SelectedTrain.Line.Color;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Style := psClear;
    Pos := LastMousePos;
    Angle := 0;

    Canvas.Polygon([
      RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y - TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y + TrainSize div 3), Angle),
      RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y + TrainSize div 3), Angle)
    ]);
  end;

  // Show grabbed terminal by mouse
  if SelectedTerminal then begin
    Canvas.Brush.Color := Colors.Text; //SelectedTrain.Line.Color;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Style := psClear;
    Pos := LastMousePos;
    Angle := 0;

    Canvas.Ellipse(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 2,
      Pos.X + TrainSize div 2, Pos.Y + TrainSize div 2);
  end;
end;

procedure TEngine.Tick;
var
  Passenger: TMetroPassenger;
  MapStation: TMapStation;
begin
  if State = gsRunning then begin
    FTime := FTime + (Now - LastTickTime) / OneSecond * TimePerSecond;
    Redraw; // Redraw on every tick because engine time is changed so clock should be redrawn

    // Add new trains
    if (Time - LastNewWeekTime) > NewTrainPeriod then begin
      LastNewWeekTime := Time;
      Inc(Week);
      State := gsNewWeek;
      Redraw;
    end;

    // Add new shape
    if (Time - LastNewShapeTime) > NewShapePeriod then begin
      LastNewShapeTime := Time;
      if ShapeCount <= Integer(High(TStationShape)) then Inc(ShapeCount);
      Redraw;
    end;

    // Add new station
    if (Time - LastNewStationTime) > NewStationPeriod then begin
      LastNewStationTime := Time;
      Stations.AddNew;
      ResizeView(False);
      Redraw;
    end;

    // Add new passengers
    if (Time - LastNewPassengerTime) > NewPassengerPeriod then begin
      LastNewPassengerTime := Time;
      for MapStation in Stations do
      with MapStation do
      if Random < NewPassengerProbability then begin
        Passenger := Self.Passengers.AddNew;
        Passenger.Shape := TStationShape(Random(Integer(ShapeCount)));
        Passengers.Add(Passenger);

        // Passenger is not allowed to have same shape
        while (Passenger.Shape = Shape) or
        not (Passenger.Shape in GetExistStationShapes) do
          Passenger.Shape := TStationShape((Integer(Passenger.Shape) + 1) mod Integer(ShapeCount));
        Redraw;
      end;
    end;

    // Check station passenger overload state
    for MapStation in Stations do
    with MapStation do begin
      if Passengers.Count > GetMaxPassengers then begin
        OverloadDuration := OverloadDuration + (FTime - FLastTime);
        if OverloadDuration > MaxPassengersOveloadTime then
          OverloadDuration := MaxPassengersOveloadTime;
        if OverloadDuration < MaxPassengersOveloadTime then Redraw;
      end;
      if Passengers.Count <= GetMaxPassengers then begin
        if OverloadDuration > 0 then Redraw;
        OverloadDuration := OverloadDuration - (FTime - FLastTime);
        if OverloadDuration < 0 then begin
          OverloadDuration := 0;
        end;
      end;
    end;

    TrainsMovement;

    // Game over
    for MapStation in Stations do
    with MapStation do begin
      if OverloadDuration >= MaxPassengersOveloadTime then begin
        State := gsGameOver;
        Redraw;
      end;
    end;

    if Assigned(City) and (ServedPassengerCount >= City.PassengersCountToUnlock) then begin
      State := gsSuccess;
      Redraw;
    end;
  end;
  LastTickTime := Now;
  FLastTime := FTime;
end;

procedure TEngine.MouseMove(Position: TPoint);
var
  FocusedStation: TMapStation;
  Line: TMetroLine;
  LineStationDown: TLineStation;
  LineStationUp: TLineStation;
  CurrentTrackPoint: TTrackPoint;
begin
  if State = gsMenu then begin
    Menu.MouseMove(Position);
    Redraw;
  end;

  if Assigned(SelectedTrain) or Assigned(SelectedCarriage) or
    SelectedTerminal then Redraw;

  LastMousePos := Position;
  if MouseHold then begin
    FocusedStation := GetStationOnPos(View.PointDestToSrc(Position));
    Line := nil;
    if Assigned(TrackStationDown) then begin
      Line := TMetroLine(TrackStationDown.Track.Owner);
      Redraw;
    end;
    if Assigned(TrackStationUp) then begin
      Line := TMetroLine(TrackStationUp.Track.Owner);
      Redraw;
    end;
    if Assigned(Line) and not Assigned(LastFocusedStation) and Assigned(FocusedStation) then begin
      if Assigned(TrackStationDown) and (TLineStation(TrackStationDown.OwnerPoint).MapStation = FocusedStation) then begin
        // Disconnect down
        CurrentTrackPoint := TrackStationDown;
        TrackStationDown := TrackStationDown.GetDown;
        Line.DisconnectStation(TLineStation(CurrentTrackPoint.OwnerPoint));
      end else
      if Assigned(TrackStationUp) and (TLineStation(TrackStationUp.OwnerPoint).MapStation = FocusedStation) then begin
        // Disconnect up
        CurrentTrackPoint := TrackStationUp;
        if Assigned(TrackStationUp) then
          TrackStationUp := TrackStationUp.GetUp;
        Line.DisconnectStation(TLineStation(CurrentTrackPoint.OwnerPoint));
      end else
      if Assigned(Line) and ((not Line.IsCircular) or ((TrackStationDown <> nil) and (TrackStationUp <> nil))) and
      ((Line.LineStations.SearchMapStation(FocusedStation) = nil) or
      ((Line.LineStations.Count > 0) and
      ((Line.LineStations.First.MapStation = FocusedStation) or
      (Line.LineStations.Last.MapStation = FocusedStation)) and
      ((TrackStationDown = nil) or (TrackStationUp = nil)) and
      (not Line.IsCircular))) then begin
        if Assigned(TrackStationDown) then LineStationDown := TLineStation(TrackStationDown.OwnerPoint)
          else LineStationDown := nil;
        if Assigned(TrackStationUp) then LineStationUp := TLineStation(TrackStationUp.OwnerPoint)
          else LineStationUp := nil;
        Line.ConnectStation(FocusedStation, LineStationDown, LineStationUp);
        if Assigned(TrackStationDown) then TrackStationDown := TrackStationDown.GetUp
          else if Assigned(TrackStationUp) then TrackStationUp := TrackStationUp.GetDown;
      end;
    end;
    LastFocusedStation := FocusedStation;
  end;
end;

procedure TEngine.MouseUp(Button: TMouseButton; Position: TPoint);
var
  I: Integer;
  FocusedTrack: TTrackLink;
  FocusedTrain: TMetroTrain;
  FocusedStation: TMapStation;
  Intersect: TPoint;
begin
  if Button = mbLeft then begin
    if State = gsMenu then begin
      Menu.MouseUp(Button, Position);
      Redraw;
    end else
    if State in [gsGameOver, gsSuccess] then begin
      ButtonBack.MouseUp(Position);
    end else
    if State = gsNewWeek then begin
      ImageNewTrain.MouseUp(Position);
    end else
    if State = gsNewImprovement then begin
      ImageNewImprovement1.MouseUp(Position);
      ImageNewImprovement2.MouseUp(Position);
    end
    else
    if State in [gsRunning, gsPaused] then begin
      ButtonBack.MouseUp(Position);
      ImagePause.MouseUp(Position);
      ImagePlay.MouseUp(Position);
      ImageFastForward.MouseUp(Position);
      Redraw;

      // Place selected train if focused track
      if Assigned(SelectedTrain) then begin
        SelectedTrain.TargetStation := nil;
        SelectedTrain.TrackPosition.BaseTrackPoint := nil;
        if Assigned(SelectedTrain.Line) then begin
          SelectedTrain.Line.Trains.Remove(SelectedTrain);
          SelectedTrain.Line := nil;

          // Remove train carriages
          for I := SelectedTrain.Carriages.Count - 1 downto 0 do begin
            SelectedTrain.Carriages[I].Train := nil;
            SelectedTrain.Carriages.Delete(I);
          end;
        end;
        FocusedTrack := GetTrackOnPos(View.PointDestToSrc(Position), Intersect);
        if Assigned(FocusedTrack.Points[0]) then begin
          SelectedTrain.Line := TMetroLine(FocusedTrack.Points[0].Track.Owner);
          SelectedTrain.Line.Trains.Add(SelectedTrain);
          SelectedTrain.TrackPosition.BaseTrackPoint := FocusedTrack.Points[0];
          SelectedTrain.TrackPosition.RelPos := Distance(FocusedTrack.Points[0].Position,
            Intersect);
          SelectedTrain.Direction := GrabbedTrainDirection;
          SelectedTrain.FindTargetStation;
          SelectedTrain.LastTrainMoveTime := Time;
        end;
        FocusedTrack.Free;

        LastGrabbedTrain := nil;
      end;

      // Place selected carriage if focused train
      if Assigned(SelectedCarriage) then begin
        if Assigned(SelectedCarriage.Train) then begin
          SelectedCarriage.Train.Carriages.Remove(SelectedCarriage);
          SelectedCarriage.Train := nil;
        end;
        FocusedTrain := GetTrainOnPos(View.PointDestToSrc(Position));
        if Assigned(FocusedTrain) then begin
          SelectedCarriage.Train := FocusedTrain;
          FocusedTrain.Carriages.Add(SelectedCarriage);
        end;
      end;

      if SelectedTerminal then begin
        FocusedStation := GetStationOnPos(View.PointDestToSrc(Position));
        if Assigned(FocusedStation) and not FocusedStation.IsTerminal then begin
          FocusedStation.IsTerminal := True;
          Dec(AvailableTerminals);
          Redraw;
        end;
        SelectedTerminal := False;
      end;

      // Line color selection
      for I := 0 to Lines.Count - 1 do
        if Distance(Point(View.DestRect.Right div 2 - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
          View.DestRect.Bottom - LineColorsDist), Position) < 20 then begin
            SelectedLine := Lines[I];
            Exit;
          end;

      // Remove single line station on line
      if Assigned(TrackStationDown) and (TMetroLine(TrackStationDown.Track.Owner).LineStations.Count = 1) then begin
        TMetroLine(TrackStationDown.Track.Owner).DisconnectStation(
          TMetroLine(TrackStationDown.Track.Owner).LineStations.First);
      end;
      if Assigned(TrackStationUp) and (TMetroLine(TrackStationUp.Track.Owner).LineStations.Count = 1) then begin
        TMetroLine(TrackStationUp.Track.Owner).DisconnectStation(
          TMetroLine(TrackStationUp.Track.Owner).LineStations.First);
      end;
    end;
  end else
  if Button = mbRight then begin
    SelectedLine := nil;
  end;
  MouseHold := False;
  TrackStationDown := nil;
  TrackStationUp := nil;
  SelectedTrain := nil;
  SelectedCarriage := nil;
  Redraw;
end;

procedure TEngine.MouseDown(Button: TMouseButton; Position: TPoint);
var
  Station: TMapStation;
  NewLine: TMetroLine;
  Track: TTrackLink;
  NewIndex: Integer;
  Intersection: TPoint;
begin
  if (Button = mbLeft) then begin
    if State in [gsRunning, gsPaused] then begin
      MouseHold := True;
      LastFocusedStation := nil;
      Redraw;

      // Train selection
      SelectedTrain := GetTrainOnPos(View.PointDestToSrc(Position));
      if Assigned(SelectedTrain) then begin
        Exit;
      end;

      // Carriage selection
      SelectedCarriage := GetCarriageOnPos(View.PointDestToSrc(Position));
      if Assigned(SelectedCarriage) then begin
        Exit;
      end;

      // Select unused train
      if (Distance(Position, ImageLocomotive.Bounds.CenterPoint) < 30) and
      (Trains.GetUnusedCount > 0) then begin
        SelectedTrain := Trains.GetUnused;
        Exit;
      end;

      // Select unused carriage
      if CarriageCountVisible and (Distance(Position, ImageCarriage.Bounds.CenterPoint) < 30) and
      (Carriages.GetUnusedCount > 0) then begin
        SelectedCarriage := Carriages.GetUnused;
        Exit;
      end;

      // Select unused carriage
      if TerminalCountVisible and (Distance(Position, ImageTerminal.Bounds.CenterPoint) < 30) and
      (AvailableTerminals > 0) then begin
        SelectedTerminal := True;
        Exit;
      end;

      // New track creation from selected station as start
      Station := GetStationOnPos(View.PointDestToSrc(Position));
      if Assigned(Station) then begin
        NewLine := GetSelectedOrUnusedMetroLine;
        if Assigned(NewLine) then begin
          NewLine.ConnectStation(Station, nil, nil);
          TrackStationDown := NewLine.Track.Points.Last;
          TrackStationUp := nil;
          LastFocusedStation := Station;
          SelectedLine := NewLine;
          Exit;
        end;
      end;

      // Line selection
      Track := GetTrackOnPos(View.PointDestToSrc(Position), Intersection);
      if Assigned(Track) and Assigned(Track.Points[0]) and Assigned(Track.Points[1]) then begin
        SelectedLine := TMetroLine(Track.Points[0].Track.Owner);

        TrackStationDown := Track.Points[0];
        NewIndex := TrackStationDown.Track.Points.IndexOf(TrackStationDown);
        while Assigned(TrackStationDown) and (not Assigned(TrackStationDown.OwnerPoint)) do begin
          NewIndex := NewIndex - 1;
          if NewIndex >= 0 then TrackStationDown := TrackStationDown.Track.Points[NewIndex]
            else TrackStationDown := nil;
        end;
        TrackStationUp := Track.Points[1];
        NewIndex := TrackStationUp.Track.Points.IndexOf(TrackStationDown);
        while Assigned(TrackStationUp) and (not Assigned(TrackStationUp.OwnerPoint)) do begin
          NewIndex := NewIndex + 1;
          if NewIndex < TrackStationUp.Track.Points.Count then
            TrackStationUp := TrackStationUp.Track.Points[NewIndex]
            else TrackStationUp := nil;
        end;
        Track.Free;
        Exit;
      end;
      if Assigned(Track) then Track.Free;
    end;
  end;
end;

procedure TEngine.KeyUp(Key: Word);
const
  KeyEsc = 27;
{$IFDEF DEBUG}
  KeyF2 = 113;
  KeyF3 = 114;
  KeyF4 = 115;
  KeyF5 = 116;
  KeyF6 = 117;
  KeyF7 = 118;
  KeyF8 = 119;
{$ENDIF}
  KeyT = 84;
  KeyC = 67;
  KeyF = 70;
  KeyP = 80;
begin
  if Key = KeyEsc then begin
    if State = gsMenu then begin
      if Assigned(Menu.OnExit) then
        Menu.OnExit(nil);
    end else begin
      ButtonBackClick(nil);
    end;
  end;
  if State in [gsRunning, gsPaused] then begin
    if Key = KeyT then begin
      if Trains.GetUnusedCount > 0 then begin
        SelectedTrain := Trains.GetUnused;
        Redraw;
      end;
    end else
    if Key = KeyC then begin
      if Carriages.GetUnusedCount > 0 then begin
        SelectedCarriage := Carriages.GetUnused;
        Redraw;
      end;
    end else
    if Key = KeyF then begin
      ButtonFastForward(Self);
    end else
    if Key = KeyP then begin
      ButtonPlay(Self);
    end;
  end;
  {$IFDEF DEBUG}
  if State in [gsRunning, gsPaused] then begin
    if Key = KeyF2 then begin
      State := gsGameOver;
      Redraw;
    end;
    if Key = KeyF3 then begin
      Trains.AddNew;
      Redraw;
    end;
    if Key = KeyF4 then begin
      Carriages.AddNew;
      Redraw;
    end;
    if Key = KeyF5 then begin
      State := gsNewWeek;
      Redraw;
    end;
    if Key = KeyF6 then begin
      Stations.AddNew;
      ResizeView(False);
      Redraw;
    end else
    if Key = KeyF7 then begin
      State := gsSuccess;
      Redraw;
    end else
    if Key = KeyF8 then begin
      Inc(AvailableTerminals);
      Redraw;
    end;
  end;
  {$ENDIF}
end;

procedure TEngine.MainMenu;
begin
  State := gsMenu;
  Redraw;
end;

procedure TEngine.Clear;
begin
  CarriageCountVisible := False;
  TerminalCountVisible := False;
  AvailableTerminals := 0;
  ServedPassengerCount := 0;
  Week := 1;
  Trains.Clear;
  Passengers.Clear;
  Carriages.Clear;
  Lines.Clear;
  Stations.Clear;
  View.SourceRect := Bounds(0, 0, 0, 0);
  SelectedLine := nil;
  SelectedTrain := nil;
  SelectedCarriage := nil;
end;

procedure TEngine.NewGame;
var
  NewTrain: TMetroTrain;
  I: Integer;
  NewStation: TMapStation;
  InitialStationCount: Integer;
  InitialLineCount: Integer;
begin
  Clear;
  if Assigned(City) then begin
    LineColors := City.LineColors;
    InitialLineCount := City.InitialLineCount;
  end else begin
    LineColors := [clBlue, clRed, clDarkYellow, clGreen, clPurple, clGray,
      clOrange, clBrown, clCyan];
    InitialLineCount := 3;
  end;
  AvailableImprovements := [miCarriage, miLine, miTerminal];
  ShapeCount := 3;

  // Start with 3 stations with each different shape
  InitialStationCount := 3;
  for I := 0 to InitialStationCount - 1 do begin
    NewStation := Stations.AddNew;
    if I = 0 then NewStation.Shape := ssSquare
    else if I = 1 then NewStation.Shape := ssCircle
    else if I = 2 then NewStation.Shape := ssTriangle;
  end;

  for I := 0 to InitialLineCount - 1 do begin
    Lines.AddNew(LineColors[Lines.Count]);
    NewTrain := TMetroTrain.Create;
    Trains.Add(NewTrain);
  end;

  ResizeView(True);

  FTime := 0;
  FLastTime := 0;
  LastNewStationTime := Time;
  LastNewPassengerTime := Time;
  LastNewWeekTime := Time;
  LastNewShapeTime := Time;
  LastTickTime := Now;
  State := gsRunning;
  Redraw;
end;

procedure TEngine.Redraw;
begin
  RedrawPending := True;
end;

procedure TEngine.LoadFromRegistry;
begin
  with TRegistryEx.Create do
  try
    CurrentContext := RegistryContext;
    DarkMode := ReadBoolWithDefault('DarkMode', False);
    HighestServedPassengerCount := ReadIntegerWithDefault('HighestPassengers', 0);
    HighestServedDaysCount := ReadIntegerWithDefault('HighestDays', 0);
    Cities.LoadFromRegistry(TRegistryContext.Create(RegistryContext.RootKey, RegistryContext.Key + '\Cities'));
  finally
    Free;
  end;
end;

procedure TEngine.SaveToRegistry;
begin
  with TRegistryEx.Create do
  try
    CurrentContext := RegistryContext;

    WriteBool('DarkMode', DarkMode);
    WriteInteger('HighestPassengers', HighestServedPassengerCount);
    WriteInteger('HighestDays', HighestServedDaysCount);
    Cities.SaveToRegistry(TRegistryContext.Create(RegistryContext.RootKey, RegistryContext.Key + '\Cities'));
  finally
    Free;
  end;
end;

procedure TEngine.LoadFromFile(FileName: string);
var
  NewNode: TDOMNode;
  Doc: TXMLDocument;
  RootNode: TDOMNode;
begin
  ReadXMLFile(Doc, FileName);
  Clear;
  with Doc do try
    if Doc.DocumentElement.NodeName <> 'BigMetroGame' then
      raise Exception.Create(SWrongFileFormat);
    RootNode := Doc.DocumentElement;
    with RootNode do begin
      ServedPassengerCount := ReadInteger(RootNode, 'ServedPassengerCount', ServedPassengerCount);

      NewNode := FindNode('Stations');
      Stations.LoadFromXmlNode(NewNode);

      NewNode := FindNode('Lines');
      Lines.LoadFromXmlNode(NewNode);

      NewNode := FindNode('Trains');
      Trains.LoadFromXmlNode(NewNode);
    end;
  finally
    FreeAndNil(Doc);
  end;
end;

procedure TEngine.SaveToFile(FileName: string);
var
  NewNode: TDOMNode;
  Doc: TXMLDocument;
  RootNode: TDOMNode;
  I: Integer;
begin
  Doc := TXMLDocument.Create;
  with Doc do try
    RootNode := CreateElement('BigMetroGame');
    AppendChild(RootNode);
    with RootNode do begin
      WriteInteger(RootNode, 'ServedPassengerCount', ServedPassengerCount);

      for I := 0 to Lines.Count - 1 do
        Lines[I].Id := I + 1;

      NewNode := OwnerDocument.CreateElement('Stations');
      AppendChild(NewNode);
      Stations.SaveToXmlNode(NewNode);

      NewNode := OwnerDocument.CreateElement('Lines');
      AppendChild(NewNode);
      Lines.SaveToXmlNode(NewNode);

      NewNode := OwnerDocument.CreateElement('Trains');
      AppendChild(NewNode);
      Trains.SaveToXmlNode(NewNode);
    end;
    if ExtractFileDir(FileName) <> '' then
      ForceDirectories(ExtractFileDir(FileName));
    WriteXMLFile(Doc, FileName);
  finally
    FreeAndNil(Doc);
  end;
end;

constructor TEngine.Create(AOwner: TComponent);
begin
  inherited;
  ImprovementImageSize := ScaleX(64, 96);
  IconSize := ScaleX(32, 96);
  TimePerSecond := TimePerSecondNormal;
  ButtonBack := TImage.Create;
  ButtonBack.OnClick := ButtonBackClick;
  ButtonBack.Bounds := Bounds(0, 0, ScaleX(80, 96), ScaleY(80, 96));
  Cities := TCities.Create;
  InitCities;
  MenuMain := TMenu.Create;
  MenuOptions := TMenu.Create;
  MenuCareer := TMenu.Create;
  MenuGame := TMenu.Create;
  Menu := MenuMain;
  InitMenus;
  Stations := TMapStations.Create;
  Stations.Engine := Self;
  Lines := TMetroLines.Create;
  Lines.Engine := Self;
  Passengers := TMetroPassengers.Create;
  Map := TMap.Create;
  View := TView.Create;
  Trains := TMetroTrains.Create;
  Carriages := TMetroCarriages.Create;
  ImageTunnel := TImage.Create;
  ImageLine := TImage.Create;
  ImageTerminal := TImage.Create;
  ImagePassenger := TImage.Create;
  ImageLocomotive := TImage.Create;
  ImageCarriage := TImage.Create;
  ImagePlay := TImage.Create;
  ImagePlay.OnClick := ButtonPlay;
  ImagePause := TImage.Create;
  ImagePause.OnClick := ButtonPause;
  ImageFastForward := TImage.Create;
  ImageFastForward.OnClick := ButtonFastForward;
  ImageAchievement := TImage.Create;
  ImageNewTrain := TImage.Create;
  ImageNewTrain.OnClick := ButtonNewTrain;
  ImageNewImprovement1 := TImage.Create;
  ImageNewImprovement1.OnClick := ButtonNewImprovement1;
  ImageNewImprovement2 := TImage.Create;
  ImageNewImprovement2.OnClick := ButtonNewImprovement2;
  MetaCanvas := TMetaCanvas.Create;
  Colors.Init(FDarkMode);
end;

destructor TEngine.Destroy;
begin
  FreeAndNil(MetaCanvas);
  FreeAndNil(Trains);
  FreeAndNil(Carriages);
  FreeAndNil(ImageTunnel);
  FreeAndNil(ImageTerminal);
  FreeAndNil(ImageLine);
  FreeAndNil(ImageNewImprovement1);
  FreeAndNil(ImageNewImprovement2);
  FreeAndNil(ImageNewTrain);
  FreeAndNil(ImagePlay);
  FreeAndNil(ImageFastForward);
  FreeAndNil(ImageAchievement);
  FreeAndNil(ImagePause);
  FreeAndNil(ImageCarriage);
  FreeAndNil(ImageLocomotive);
  FreeAndNil(ImagePassenger);
  FreeAndNil(View);
  FreeAndNil(Map);
  FreeAndNil(Passengers);
  FreeAndNil(Stations);
  FreeAndNil(Lines);
  FreeAndNil(MenuMain);
  FreeAndNil(MenuOptions);
  FreeAndNil(MenuCareer);
  FreeAndNil(MenuGame);
  FreeAndNil(Cities);
  FreeAndNil(ButtonBack);
  inherited;
end;

procedure TEngine.Paint(Canvas: TCanvas; CanvasSize: TPoint);
begin
  MetaCanvas.Size := Point(Canvas.Width, Canvas.Height);
  MetaCanvas.Reset;

  DrawStationPassengerOverload(MetaCanvas);
  DrawLines(MetaCanvas);
  DrawTrains(MetaCanvas);
  DrawStations(MetaCanvas);

  // MainMenu background
  Canvas.Brush.Color := Colors.Background;
  Canvas.Brush.Style := bsSolid;
  Canvas.Clear;

  MetaCanvas.Move(Point(-View.SourceRect.Left, -View.SourceRect.Top));
  MetaCanvas.Zoom(View.Zoom);

  // Draw meta canvas to real target canvas
  MetaCanvas.DrawTo(Canvas);

  if State = gsGameOver then
  begin
    DrawGameOver(Canvas, CanvasSize);
    DrawGameControls(Canvas, CanvasSize);
  end else
  if State = gsSuccess then
  begin
    DrawSuccess(Canvas, CanvasSize);
    DrawGameControls(Canvas, CanvasSize);
  end else
  if State = gsMenu then begin
    Menu.Paint(Canvas, CanvasSize);
  end else
  if State = gsNewWeek then begin
    DrawNewWeek(Canvas, CanvasSize);
  end else
  if State = gsNewImprovement then begin
    DrawNewImprovement(Canvas, CanvasSize);
  end else
  if State in [gsRunning, gsPaused] then begin
    DrawGameControls(Canvas, CanvasSize);
    DrawGrabbed(Canvas, CanvasSize);
  end;

  RedrawPending := False;
end;

end.

